"""
This module contains a collection of functions intended for simple data anlysis and preprocessing of the data.
These are basically in continuous grow from the possible use cases in which :ref:`index:gwdama` finds applications. 

Some of these functions are also provided as *methods* for the :ref:`dataset:Dataset` class.
"""
import numpy as np
import scipy
#from scipy import signal
from scipy.interpolate import interp1d
from scipy.signal import get_window, argrelmin, argrelmax, spectrogram, welch, decimate,  windows, freqz, periodogram
from multiprocessing import Pool, cpu_count
from functools import partial

def whiten(strain, interp_psd, dt, phase_shift=0, time_shift=0):
    """Function that returns the whitened strain data given the PSD and sample rate, also optionally applying a phase
    shift and time shift.

    Parameters
    ----------
    strain : array_like
        a time series, such as the strain data
    interp_psd : interpolating function
        function to take in freqs and output the average power at that freq 
    dt : float
        sample time interval of data
    phase_shift : float, optional
        phase shift to apply to whitened data
    time_shift : float, optional
        time shift to apply to whitened data (s)
    
    Returns
    -------
    : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of whitened strain data
    """
    Nt = len(strain)
    # take the fourier transform of the data
    freqs = np.fft.rfftfreq(Nt, dt)

    # whitening: transform to freq domain, divide by square root of psd, then
    # transform back, taking care to get normalization right.
    hf = np.fft.rfft(strain)
    
    # apply time and phase shift
    hf = hf * np.exp(-1.j * 2 * np.pi * time_shift * freqs - 1.j * phase_shift)
    norm = 1./np.sqrt(1./(dt*2))
    white_hf = hf / np.sqrt(interp_psd(freqs)) * norm
    white_ht = np.fft.irfft(white_hf, n=Nt)
    return white_ht

def bandpass(tsdata, fband, fs):
    """Bandpasses strain data using a butterworth filter.
    
    Parameters
    ----------
    tsdata :  array_likje
        Timeseries data to bandpass
    fband : (float,float)
        Low and high-pass filter values to use
    fs : float
        Sample rate of data
    
    Returns
    -------
    : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Array of bandpassed timeseries data
    """
    from scipy.signal import butter, filtfilt
    bb, ab = butter(4, [fband[0]*2./fs, fband[1]*2./fs], btype='band')
    normalization = np.sqrt((fband[1]-fband[0])/(fs/2))
    data_bp = filtfilt(bb, ab, tsdata) / normalization
    return data_bp

def next_power_of_2(x):
    """
    Silly functions to obtain the "next power of 2" for any given
    number ``x``. It is useful for padding before computing the FFTs needed for spectral estimations, such as for the Welch's method.
    This is computed by means of the `bit_length method <https://docs.python.org/3/library/stdtypes.html#int.bit_length>`_ of ints, corresponding to the number of bits necessary to represent an integer in binary, excluding the sign and leading zeros
    
    Parameters
    ----------
    x : int
        Number, for example the length of an array
    Returns
    -------
    n : int
        equals to 1 if ``x==0``, the next power of two of ``x`` otherwise
        
    """   
    return 1 if x == 0 else 2**(x - 1).bit_length()


def taper(data, fs=1, side='leftright', duration=None, nsamples=None, window=('tukey',0.25)):
    """Taper the edges of this datset smoothly to zero. The method automatically tapers from the second stationary point (local maximum or minimum) on the specified side of the input. However, the method will never taper more than half the full width of the data, and will fail if there are no stationary points.
        
    Parameters
    ----------
    fs   : float, optional
        sampling frequency of the data. If not specified this is assumed to be 1
    side : str, optional
        the side of the data to taper, must be one of ``'left'``,
        ``'right'``, or ``'leftright'``
    duration : float, optional
        the duration of time to taper, will override ``nsamples`` if both are provided as arguments
    nsamples : int, optional
        the number of samples to taper, will be overridden by ``duration`` if both are provided as arguments
    window : string, float, or tuple
        The type of window to create. Accepts everything that can be passed to `scipy.signal.get_window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.get_window.html>`_
       
    Returns
    -------
    out : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        a copy of data tapered at one or both ends
    Raises
    ------
    ValueError
        if ``side`` is not one of ``('left', 'right', 'leftright')``
    """
    # check window properties
    if side not in ('left', 'right', 'leftright'):
        raise ValueError("side must be one of 'left', 'right', or 'leftright'")
        
    if duration:
        nsamples = int(duration * fs)
        
    out = data.copy()
    # if a duration or number of samples is not specified, automatically
    # identify the second stationary point away from each boundary,
    # else default to half the data width
    nleft, nright = 0, 0
    if not nsamples:
        mini, = argrelmin(out)
        maxi, = argrelmax(out)
        
    if 'left' in side:
        nleft = nsamples or max(mini[0], maxi[0])
        nleft = min(nleft, int(data.size/2))
    if 'right' in side:
        nright = nsamples or out.size - min(mini[-1], maxi[-1])
        nright = min(nright, int(data.size/2))
    
    # Define the window
    win_len = (nleft or 0)+(nright or 0)  
    
    win_func = get_window(window=window, Nx=win_len)
    #val("scipy.signal.windows.{}({}, {})".format(window, win_len, win_par))
    
    out[:(nleft or 0)] *= win_func[:(nleft or 0)]
    out[-(nright or 0):] *= win_func[-(nright or 0):]

    return out


def avg_spectro(x, stride, fftlength=None, overlap=None, t0=None, fs=None, window='blackman', average='mean', **psdkwgs):
    """
    Function to compute the averaged spectrogram of a time series ``x``. This is based on `scipy.signal.spectrogram <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.spectrogram.html>`_ function, with the key parameters redefined in units of seconds instead of points. This function computes consecutive Fourier transforms of the data. The result is then averaged (approximately) every ``stride`` seconds, according to the selected ``average`` method (classical choices are ``'mean'`` and ``'median'``). Notice that, depending on the choice of ``fftlength``, ``overlap``, and the origninal sampling frequency of the data ``fs``, the output has only approximatively a spacing of ``stride`` seconds.  This is done to avoid data loss, when the number of points in each non-overlapping FFT is not a sub-multiple of the point included in every stride. If you prefer to have exactly ``stride`` seconds spaced estimates, with consequent data loss, make use of the other ``multi_spectro`` method.
    
    Parameters
    ----------
    x : array_like
        Time series of measurement values
    stride : float
        (approximate) number of seconds in single column of spectrogram
    fftlength : float, optional
        length in seconds of the FFT to use to estimate the PSD. Default to the value of ``stride`` if not provided
    overlap :   float, optional
        overlap in seconds of the segments to use for computing the PSD. Defoult is zero
    t0 : float, optional
        Epoch where to start the returned time vector
    fs : int, optional 
        sampling frequency of the dataset. Default to 1 if not provided
    window : str, float, or tuple
        The type of window to create. Accepts everything that can be passed to `scipy.signal.get_window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.get_window.html>`_
    average : str, optional
        Typical choices are ``'mean'`` or ``'median'``. In principle, every other `Average and Variance statistical function of Numpy <https://numpy.org/doc/stable/reference/routines.statistics.html>`_ that can take an array as a single parameter is fine as well (but it is not guaranteed that it'll necessarily make sense, though). E.g.: ``'min'`` and ``'max'`` are legitimate choices (but it is up to you give them a meaningful interpretation)
    **psdkwgs : optional
        other parameters to be passed to `scipy.signal.spectrogram <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.spectrogram.html>`_
    
    Returns
    -------
    t : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of segment times
    f : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of sample frequencies
    Sxx : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Spectrogram of x. By default, the last axis of Sxx corresponds to the segment times
    
    See Also
    --------
    gwdama.preprocessing.multi_spectro
    """
    if not fs:
        fs = 1
    
    # fftlength
    if fftlength is None: 
        nperseg = int(stride*fs) 
    else:
        nperseg = int(fftlength*fs)
    
    # noverlap
    if overlap is None:
        noverlap = 0
    else:
        noverlap = int(overlap*fs)
    
    # nfft, if not specified, is set equal to the next power of two of nperseg (zero padding)
    nfft=psdkwgs.get('nfft', 2**(nperseg - 1).bit_length())
    psdkwgs.update(dict(window=window,
                        nperseg=nperseg,
                        nfft=nfft,
                        fs=fs))   
       
    ff, tt, Sxx = spectrogram(x, **psdkwgs)
    if t0:
        tt +=t0
        
    nstride = sum(tt-tt[0]<stride)
    nmax = (len(tt)//nstride *nstride)
    strides = len(tt)//nstride
    B = Sxx[:,:nmax].reshape(Sxx.shape[0], strides,-1)
    C = eval(f"np.{average}(B,-1)")
    return tt[:nmax:nstride], ff, C


def multi_spectro(x, stride, fftlength=None, overlap=None, t0=None, fs=None, window='blackman', average='mean', nproc=1, **psdkwgs):
    """
    Function to compute a specttrogram  of a time series ``x`` over multiple processes.
    This is based on `scipy.signal.welch <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_ function, with the key parameters redefined in units of seconds instead of points. This function estimates over different processes the PSD of the data every ``stride`` seconds by means of some modified versions of the Welch's method, that is averaging windowed (and possibly zero padded) FFTs of the data over overlapping segments. Valid averaging methods are the usual ``'mean'`` and ``'median'``, more robust to outliers in the data (such as glitches and other non-stastioarities).
    Notice that, depending on the choice of ``fftlength``, ``overlap``, and the origninal sampling frequency of the data ``fs``, the PSD may not consider the last data in each segment. This happens typically happens when ``fftlength-overlap`` is not a submultiple of the ``stride``. 
    If you prefer to avoid losing data in each segments, but speed speed and time spacing at exactly the value of ``stride`` is not required, the use of the other ``avg_spectro`` function is preferrable. This one is usually faster, though, making use of multiple processes.
    
    Parameters
    ----------
    x : array_like
        Time series of measurement values
    stride : float
        exact number of seconds for each PSD estimate, and therefore in each single column of the spectrogram
    fftlength : float, optional
        length in seconds of the FFT to use to estimate the PSD. Default to the value of ``stride`` if not provided
    overlap :   float, optional
        overlap in seconds of the segments to use for computing the PSD. Defoult is zero
    t0 : float, optional
        Epoch where to start the returned time vector
    fs : int, optional 
        sampling frequency of the dataset. Default to 1 if not provided
    window : str, float, or tuple
        The type of window to apply to the FFTs. Accepts everything that can be passed to `scipy.signal.get_window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.get_window.html>`_. A popular and balanced choice is ``'blackman'``. Other possibilities include ``'hann'``, or ``('tukey', 0.25)``
    average : str, optional
        Available choices are ``'mean'`` or ``'median'``
    nproc : int, optional
        number of CPUs to use in parallel processing of PSDs
    **psdkwgs : optional
        other parameters to be passed to `scipy.signal.welch <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_
    
    Returns
    -------
    t : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of segment times
    f : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of sample frequencies
    Sxx : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Spectrogram of x. By default, the last axis of Sxx corresponds to the segment times
        
    See Also
    --------
    gwdama.preprocessing.avg_spectro
    """
    if not fs:
        fs = 1
    nstride = int(stride*fs)
    x_ = x[:(len(x)//nstride *nstride)].reshape(-1,nstride)
    
    time = np.arange(len(x)//nstride)*stride
    if t0:
        time = time + t0
    
    # fftlength
    if fftlength is None: 
        nperseg = nstride 
    else:
        nperseg = int(fftlength*fs)
    
    # noverlap
    if overlap is None:
        noverlap = 0
    else:
        noverlap = int(overlap*fs)
    
    nfft=psdkwgs.get('nfft', 2**(nperseg - 1).bit_length())
    psdkwgs.update(dict(window=window,
                        nperseg=nperseg,
                        nfft=nfft,
                       average=average,
                       fs=fs))   
    
    multipsd = partial(welch, **psdkwgs)
    
    with Pool(processes=nproc) as pool:
        results = np.array(pool.map(multipsd, tuple(x_)))
            
    return time, results[0,0,:], results[:,1,:].T

#####################

def decimate_recursive(data, q_factor, fs = None, **dcmkwgs):
    """
    Downsample the signal in a recursive way after applying an anti-aliasing filter. Notice that when using IIR downsampling, it is recommended to call decimate multiple times for downsampling factors higher than 13. This Decimation cannot be performed for q_factors larger than 13,                
    then it recursively performs decimation with factors lower than                    
    13 searching for multiples of 2, 3 or 5.                                                                                                                                      
    Parameters
    ----------                                                                             
    data : array_like
        The input data
    q_factor : int
        Decimation factor                                

    Returns
    -------
    decim_data : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        decimated data                                                      
    outfs : float
        New sampling frequency
        
    See Also
    --------
    `scipy.signal.decimate <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.decimate.html#scipy.signal.decimate>`_
        This funtion is a recursive verison of the one implemented in scipy
    """
    if fs is None:
        fs = 1
        
    if q_factor == 1:                                                                      
        #print("q_factor = 1, no need to decimate further...")
        return  data, fs                                                               
    if  q_factor <= 10:                                                                      
        # if decimation factor is larger than 10 it is necessary to decimate many times 
        tg = decimate(data, q_factor, **dcmkwgs)                                                   
        #print(" downsampling by: " + str(q_factor))
        return tg, fs/q_factor                                                     
    else:                                                                              
        if (q_factor%5) == 0:                                                         
            #print(" downsampling by: 5")
            tg = decimate(data,5, **dcmkwgs)                                             
            return decimate_recursive(tg,q_factor/5, fs/5, **dcmkwgs)                      
        elif (q_factor%3) == 0:                                                       
            #print(" downsampling by: 3")                                        
            tg = decimate(data,3,**dcmkwgs)
            return decimate_recursive(tg,q_factor/3, fs/3,**dcmkwgs)
        else : 
            if (q_factor%2) != 0:      
                print(" be carefull, the sampling ratio cannot be divided by 2!!") 
            #tg = scipy.signal.decimate(data,2)  
            tg = decimate(data,2,**dcmkwgs)                                              
            return decimate_recursive(tg,q_factor/2, fs/2,**dcmkwgs)                           

##### Multi-Tapering Method        
        
def _periodogram_w(win, xx, fss, **Pdict):
    """
    Auxiliary function for mtm_psd
    """
    return periodogram(x=xx, fs=fss,window=win,**Pdict)


def mtm_psd(x, fs=None, NhBW=None, Kmax=None, eigenthr=None, norm=2, nproc=1, return_eigens=False, **PSDkwgs):
    """
    Multi-Taper Method (MTM) by D.J. Thomson is a non-parametric way of computing the Power Spectral Density (PSD) of a time series averaging various periodograms obained by means of a sequence of orthonormal windows. The latter can be given by a Sequence of Discrete Prolate Spheroidal functions (DPSS), that is, the eigenvectors (*eigentapers*) solution to the variational problem of minimizing the leackage outside of a window *W* (in the frequency domain, the PSD of these functions is maximally contained inside *W*), constrained to a norm given for example by the ell-2 norm: their sum of squares is equal to one. This gives a set of independent periodogram estimates, which can be averaged over the number of windws used. The advantage of doing this is a reduction of the variance by a factor equals to the number of tapers used. This is similar to the Welch method, where a large sequence of data is divided into segments and averaged over them. 
      
    Parameters
    ----------
    x : array-like
        The time series we want to compute the Power Spectral Density (PSD) of
    fs : float, optional
        Sampling frequency of the time series ``x``
    NhBW : float, optional
        Normalized half-Bandwidth of the windows to use, which are then found to be a sequence of orthogonal Discrete Prolate Spheroidal
        functions with spectra maximally concentrated in the (normalized) interval (-NhBW,+NhBW), and subject to to the normalization condition
        given by ``norm``. The bandwidth of the windows is normalized with respect to the zoroth-frequency f0 = 1/T = fs/N, where N is the number
        of data pouints in ``x``
    Kmax : int, optional
        This parameter determines the maximum number of orthogonal windows to use. Notice that if ``eigenthr`` is set lower than 1, than a smaller number of windows can be eventually used, depending on their eigenvalues (that is, their concentration ratio inside the interval determined by ``NhBW``). In chosing this value, take also in mind that usually the first 2``NhBW``-1 eigenvalues are close to 1, while the latter are close to zero. If ``eigenthr`` is not provided, the previous condition is applied. Set ``eigenthr`` to 1 if you want all the ``Kmax`` that you set. Also, the frequency resolution of this estimate will eventually be ``fs``*``Kmax``/``len(x)``. Thake this in mind when deciding the compromize between resolution and variance
    eigenthr : float, optional
        Only windows with eigenvalue higher than this threshold are eventually used; the latter are normalized to be in (1,0). This is done in order to avoid leakage with windows that are truncated at the borders of the data segment: this will cause a characteristic 1/f behaviour of the ASD. Select a higher value of this parameter if you notice this. If not provided, the number of window to use is set such that: ``Kmax`` = 2``NhBW``-1
    norm : {2, 'approximate', 'subsample'} | None, optional
        Leave it equals to 2, corresponding to the ell-2 norm (summ of squares equal to 1). Refer to `scipy.signal.windows.dpss documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.dpss.html>`_ for more details
    nproc : int, optional
        Number of proccessors for parallel computing. If set to one (default) no parallel computing is performed, and the various periodograms are obtained through a list comprehension
    return_eigens : bool, optional
        If ``True``, the function also returns a list of eigenvalues (concentration ratios) corresponding to the previous parameter, and possibly truncated by ``eigenthr`` parameter
    **PSDkwgs : dict, optional
        Other optional arguments to compute the periodogram PSD.
    
    Returns
    -------
    : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of sample frequencies
    : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Power spectral density of ``x`` estimated with the MTM
    """
    # Pass a frequency value
    if not fs:
        fs = 1
    # Define NhBW on the base of Kmax
    if not NhBW:
        NhBW = ((Kmax or 4) + 1)/2
    # Define Kmax if not provided: if also NhBW wasn't passed, Kmax = 4
    if not Kmax:
        Kmax=int(2*NhBW -1)
    
    N = len(x)
    win_dpss, eigens = windows.dpss(N, NW=NhBW, Kmax=Kmax, return_ratios=True, norm=norm)
    # Spectrum version is a bit faster but requires an additional package. Also, no control on the normalization
    #win_dpss, eigens = dpss(N, NW, Kmax=Kmax)
    if not eigenthr:
        emax = min(len(eigens),int(2*NhBW-1))
        eigenthr=eigens[emax-1]
    win_dpss = win_dpss[eigens>=eigenthr]
    eigens = eigens[eigens>=eigenthr]
    
    PSDdict= dict(nfft=next_power_of_2(N),
                  detrend='constant', return_onesided=True,
                  scaling='density', axis=-1)
    PSDdict.update(PSDkwgs)
    
    if nproc==1:
        # List comprehension over the windows
        res = np.array([periodogram(x=x, fs=fs, window=w, **PSDdict) for w in win_dpss])
    else:
        with Pool(nproc) as p:
            res = np.array(p.map(partial(_periodogram_w, xx=x, fss=fs,**PSDdict), win_dpss))
    
    if return_eigens:
        return res[0,0], np.average(res,axis=0,weights=eigens)[1], eigens
    else:
        return res[0,0], np.average(res,axis=0,weights=eigens)[1]


def mtm_spectro(x, stride, t0=None, fs=None,
                NhBW=None, Kmax=None, eigenthr=None, norm=2, nproc=1, **PSDkwgs):
    """
    Function to compute a specttrogram  of a time series ``x`` by means of the Multi-Taper Method.
    This function integrates the ``mtm_psd()`` function with the ``multi_spectro`` function. For each segment of data of duration set with the parameter ``stride``, the PSD is estimated with the MTM method. This allows a better variance for shorter segments of data with respect to the Welch's method implemented in the other function. Notice also that this doesn't need to declare an ``fftlength`` since all the data in the stride is included in this transformation. Also, there is no loss of data.
    
    Parameters
    ----------
    x : array_like
        Time series of measurement values
    stride : float
        Exact number of seconds for each PSD estimate, and therefore in each single column of the spectrogram
    t0 : float, optional
        Epoch where to start the returned time vector
    fs : int, optional 
        sampling frequency of the dataset. Default to 1 if not provided
    NhBW : float, optional
        Normalized half-Bandwidth of the windows to use
    Kmax : int, optional
        Maximum number of orthogonal windows to use
    eigenthr : float, optional
        Only windows with eigenvalue higher than this threshold are eventually used. If not provided, the number of window to use is set such that: ``Kmax`` = 2``NhBW``-1
    norm : {2, 'approximate', 'subsample'} | None, optional
        Leave it equals to 2, corresponding to the ell-2 norm (summ of squares equal to 1). Refer to `scipy.signal.windows.dpss documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.dpss.html>`_ for more details
    nproc : int, optional
        Number of proccessors for parallel computing. If set to one (default) no parallel computing is performed, and the various periodograms are obtained through a list comprehension
    **PSDkwgs : optional
        other parameters to be passed to `scipy.signal.welch <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_
    
    Returns
    -------
    t : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of segment times
    f : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Array of sample frequencies
    Sxx : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Spectrogram of x. By default, the last axis of Sxx corresponds to the segment times
        
    See Also
    --------
    gwdama.preprocessing.mtm_psd
        This spectrogram is obtained applying this function on multiple data segments 
    gwdama.preprocessing.multi_spectro
        This funtion is makes use of the Welch's method instead of the MTM
    """
    if not fs:
        fs = 1
    nstride = int(stride*fs)
    x_ = x[:(len(x)//nstride *nstride)].reshape(-1,nstride)
    
    time = np.arange(len(x)//nstride)*stride
    if t0:
        time = time + t0
    
    N = x_.shape[0]
    
    mtm_dict= dict(fs=fs, NhBW=NhBW, Kmax=Kmax, eigenthr=eigenthr, norm=norm, nproc=1)
    mtm_dict.update(PSDkwgs)
    
    multi_mtm = partial(mtm_psd, **mtm_dict)
    
    with Pool(processes=nproc) as pool:
        results = np.array(pool.map(multi_mtm, tuple(x_)))
            
    return time, results[0,0,:], results[:,1,:].T

# def psd

#         # number of sample for the fast fourier transform:
#         NFFT = 4 * fs
#         NOVL = -1 * NFFT / 2
#         psd_window = scipy.signal.tukey(NFFT, alpha=1./4)

#         Pxx_H1, freqs = mlab.psd(strain_H1[indxt_around], Fs=fs, NFFT=NFFT,
#                                  window=psd_window, noverlap=NOVL)
#         Pxx_L1, freqs= mlab.psd(strain_L1[indxt_around], Fs=fs, NFFT=NFFT, 
#                                 window=psd_window, noverlap=NOVL)

#         if (plot_others):
#             # smaller window if we're not doing Welch's method
#             short_indxt_away = np.where((time >= time_center - 2) & (
#                 time < time_center + 2))
#             # psd using a tukey window but no welch averaging
#             tukey_Pxx_H1, tukey_freqs = mlab.psd(
#                 strain_H1[short_indxt_away], Fs=fs, NFFT=NFFT, window=psd_window)
#             # psd with no window, no averaging
#             nowin_Pxx_H1, nowin_freqs = mlab.psd(
#                 strain_H1[short_indxt_away], Fs=fs, NFFT=NFFT, 
#                 window=mlab.window_none)

#         # We will use interpolations of the PSDs computed above for whitening:
#         psd_H1 = interp1d(freqs, Pxx_H1)
#         psd_L1 = interp1d(freqs, Pxx_L1)

#
#
#	Npt = int(infs/outfs)           # Number of samples per each segment: infs/outfs= infs*Tout
# 	Nsamples = int(len(signl)/float(Npt)) # Number of segments which the total data frame is divided into
# 	Npseg =  int(Npt/navrg)          # Number of points to include in each fft per segment
# 	Nfft = next_power_of_2(Npseg)

# 	if verbose: print("Npt: {}\nNsamples: {}\nNfft: {}".format(Npt,Nsamples,Nfft))

# 	idx = np.arange(0,Nsamples)*Npt
# 	t = np.arange(0,Nsamples)/outfs

# 	if not isinstance(bands[0],list): bands = [bands]

# 	Nbands = len(bands)

# 	# Initialize BLRMS matrix: (time vector) * (n. bands)
# 	b = np.zeros((len(t), Nbands))

	
# 	if verbose: print('PSD computation for each segment...\n frequency resolution: ',outfs*navrg,'Hz')
# 	# loop over each segment of data

# 	if nproc < 2:
# 	    for i,j in enumerate(idx):
# 	        # compute PSD
# 	        # Same as above
# 	        fr, sx = signal.welch(signl[j:j+Npt], window ='blackman', fs=infs,
# 					noverlap = Npseg/2,  nfft=Nfft, nperseg=Npseg)#, average = 'mean')


"""
Some useful pre-processing functions, like whitening a wavelte transforms
"""

from .preprocessing import whiten, taper, avg_spectro, multi_spectro, next_power_of_2, decimate_recursive, mtm_psd, mtm_spectro


import numpy as np
from multiprocessing import Pool, cpu_count
from gwpy.timeseries import TimeSeries
from gwpy.time import from_gps

# For the base classes and their extentions
import h5py

# imports related to gwdama
from .gwLogger import GWLogger

# Utilities
from gwdama.utilities import _add_method, _add_property

# For the hist and plot methods
import matplotlib.pyplot as plt
from matplotlib.figure import Figure

# To have units
import astropy.units as apu

# To raise warnings
from warnings import warn

# ----- Modified Dataset class -----

@_add_property(h5py.Dataset)
def data(self):
    """
    Returns the content of an h5py :class:`~gwdama.io.Dataset` in an easy looking way. If ``self`` has attribute ``'unit'`` and
    it is a string or a `astropy.units.Quantity <https://docs.astropy.org/en/stable/api/astropy.units.Quantity.html#astropy.units.Quantity>`_, then this is put next to the numeric value of this :class:`~gwdama.io.Dataset`. Otherwise, this method is equivalent to ``self[...]``.
    
    Returns
    ------- 
     : :py:func:`numpy.ndarray` of `astropy.units.Quantity <https://docs.astropy.org/en/stable/api/astropy.units.Quantity.html#astropy.units.Quantity>`_ 
         If a ``'unit'`` attribute is defined, the outp[ut is a `astropy.units.Quantity <https://docs.astropy.org/en/stable/api/astropy.units.Quantity.html#astropy.units.Quantity>`_, otherwise it is a :py:func:`numpy.ndarray` otherwise
    """
    if 'unit' in self.attrs:
        if isinstance(self.attrs['unit'], apu.Quantity):
            return self[...] * self.attrs['unit']
        elif isinstance(self.attrs['unit'], str):
            return self[...] * apu.Quantity(1,self.attrs['unit'])
        else:
            return self[...]
    else:
        return self[...]
            
@_add_property(h5py.Dataset)
def times(self):
    """
    Returns a Numpy array with the time vector corresponding to the curtrrent Dataset if it has 
    ``t0`` and ``sample_rate`` attributes. Raises ValueError if these attributes are not present
    
    Returns
    -------
     : :py:func:`numpy.ndarray` of `astropy.units.Quantity <https://docs.astropy.org/en/stable/api/astropy.units.Quantity.html#astropy.units.Quantity>`_
        Array of times corresponding to this Dataset
        
    Raises
    ------
    KeyError
        if the Dataset is not a time series and has no ``'t0'`` and ``'sample_rate'`` attributes
        
    """
    return (np.arange(len(self))/self.attrs['sample_rate'] + self.attrs.get('t0',0)) *apu.s

@_add_method(h5py.Dataset)
def duration(self, fs=None):
    """
    This method returns the duration in seconds of the current dataset. If the parameter ``fs`` is specified, this is the chosen *sampling frequency* of the data. Otherwise, the method attempts to access the ``sample_rate`` attribute of *self* (if it exists). If a valid ``sample_rate`` is not found, this is automatically set to 1, printing a warning message. 

    Parameters
    ----------
    fs : int, optional
        Sampling frequency of this dataset. Automatically set to 1 if not specified and a ``sample_rate`` attribute doesn't exist in the dataset
    
    Returns
    -------
    : `Astrpy.units <https://docs.astropy.org/en/stable/units/>`_
        Duration in seconds of this time series
    
    Raises
    ------
    ValueError
        if the attribute ``sample_rate`` or alternatively ``fs`` can't be converted to integer
    
    """
    # fs not given
    if fs is None:
        rate_name="sample_rate"
        # attribute exists
        if rate_name in self.attrs:
            try:
                rate = int(self.attrs[rate_name])
                return len(self.data)/rate * apu.s
            except ValueError as ve:
                print(ve, "Unrecognised format for the 'sample_rate' attribute. It can't be converted to float.")
        
        else:
            warn("WARNING!! Unrecognised attribute with the meaning of a sampling frequency. The default value will be chosen to be\
            1. Modify it passing the correct one to the 'fs' parameter of this method, or add a 'sample_rate' attribute to the dataset.")
            fs = 1
    elif isinstance(fs, (int,float)):
        return len(self.data)/fs * apu.s
    else:
        raise ValueError("Unrecognised value of the sampling frequency parameter 'fs'. Please, provide either int or float.")
                
@_add_method(h5py.Dataset)
def hist(self, closefig=True, figsize=(7,5), ax=None, **histkwgs):
    """
    Method for making a histogram of the data contained in this dataset, provided they are of numeric
    type and 1D. The output is a `Figure object of matplotlib <https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.pyplot.figure.html>`_. The title is automatically set to be the name of the channel. You can
    access and modify it later. If ``dset.data.ndim`` is higher then one, a ``ValueError`` is raised.
    
    Parameters
    ----------
    closefig : bool, optional
        choose if the returned figure object is automatically closed (``plt.close(fig)``) or not. If ``True`` (default) the figure object is closed after being created. This allows to manipulate and re-open it with the ``.reshow()`` method 
    figsize : (float, float), optional,
        Size in inches of the returned figure object. Default ``(7,5)``
    ax : Matplotlib axes object, default None
        The axes to plot the histogram on.
    **histkwgs : dict, optional
        Dictionary of all the optional parameters for the `pyplot.hist class <https://matplotlib.org/3.3.1/api/_as_gen/matplotlib.pyplot.hist.html>`_.
        
    Returns
    -------
    : `matplotlib.pyplot.figure <https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.pyplot.figure.html>`_
        Figure object of the histogram
    
    Raises
    ------
    ValueError
        If the dataset is not 1D, that is ``self.data.ndim`` > 1
    """
    if self.data.ndim != 1:
        raise ValueError("This dataset appears to be not 1D. Are you sure you whant to make an histogram out of it? If so, do it manually.")
    else:
        from gwdama.plot import make_hist
        if 'channel' in self.attrs:
            title = self.attrs['channel']
        else:
            title =''
        f = make_hist(self.data, title=title, figsize=figsize, ax=ax, **histkwgs)
        
        if ax:
            return ax
        elif closefig:
            plt.close(f)
        return f

@_add_method(h5py.Dataset)
def plot(self, figsize=(10,3), ax=None, epoch=None, scale=None, closefig=True, **axkwgs):
    """
    Plot the content of this Dataset if it is a one dimensional array. This method is meant for time series, so it seeks for attributes like the ``'sample_rate'`` (necessary) and an ``epoch`` (corresponding to the attribute ``'t0'``), which can be passed as an argument of this function.
    
    Parameters
    ----------
    figsize : (float,float), optional
        Size in inches of the returned figure object. Default ``(7,5)``  
    closefig : bool, optional
        choose if the returned figure object is automatically closed (``plt.close(fig)``) or not. If ``True`` (default) the figure object is closed after being created. This allows to manipulate and re-open it with the ``.reshow()`` method 
    ax : Matplotlib axes object, default None
        The axes to plot the histogram on.
    epoch : float, optional
        Time in seconds where to center the plot
    scale : `astropy.units <https://docs.astropy.org/en/stable/units/index.html#module-astropy.units>`_, optional
        This should be a valid `time unit <https://docs.astropy.org/en/stable/units/index.html#module-astropy.units.si>`_, for example ``'s'`` (or ``'second'``), ``'min'`` (or ``'minute'``), ``'hour'`` or ``'day'``
    **axkwgs : dict, optional
        Dictionary of all the optional parameters for the set method of matplotlib axis object.

    Returns
    -------
    : `matplotlib.pyplot.figure <https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.pyplot.figure.html>`_
        Figure object of the plot
    
    Raises
    ------
    ValueError
        If the dataset is not 1D, that is ``self.data.ndim`` > 1
    """
    if self.data.ndim != 1:
        raise ValueError("This dataset appears to be not 1D. Are you sure you whant to make an histogram out of it? If so, do it manually.")
    
    # If an axis object already exists, use it
    if ax:
        axh = ax
    else:
        fig, axh = plt.subplots(figsize=figsize)
    
    # Define x axis
    if ('t0' in self.attrs) and ('sample_rate' in self.attrs):
        xt = self.times - self.attrs.get('t0', (epoch or 0))*apu.s
    
        if not scale:
            try:
                dur = self.duration().value
                if dur//90 < 1:
                    scale = 's'
                elif dur//(90*60) <1:
                    scale = 'min'
                elif dur//(48*60*60) <1:
                    scale = 'hour'
                else:
                    scale = 'day'
                xt = xt.to(scale)
            except ValueError:
                scale = 's'            

        axh.plot(xt, self[...]);

        axh.set(xlabel=f"Time [{scale}] form {from_gps(self.attrs.get('t0', (epoch or 0))).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]} ({self.attrs.get('t0', (epoch or 0))})",ylabel=self.attrs.get('unit',''), xlim=(xt[0].value,xt[-1].value), title=self.attrs.get('channel',''))
    
    else:
        axh.plot(self[...]);
        axh.set(ylabel=self.attrs.get('unit',''), title=self.attrs.get('channel',''))
        
    axh.set(**axkwgs)
    axh.grid(True)
    
    if ax:
        return axh
    elif closefig:
        plt.close(fig)
    return fig
    
@_add_method(h5py.Dataset)
def psd(self, fftlength=None, overlap=None, fs=None, return_type='dataset', dts_key=None, **psdkwgs):
    """
    If the dataset resambles a time series (1D and with an associateded sampling frequency), this metod estimates its Power Specvtral Density (PSD) function by means of the Welch's method, or some modifications of it (median averaging instead of mean). Refer to `scipy.signal.welch <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_ for further details.
    
    Parameters
    ----------
    fftlength : float, optional 
        Length in seconds of each FFT to use to estimate the PSD. If ``None``, defoult is to use all the data lenght, *i.e.* no averaging is applied
    overlap : float, optional
        Number of seconds of overlap between each consecutive FFTs. If ``None``, no overlap is used and the parameter is set to ``0``.
    fs : int, optional
        Sampling frequency in Hz of the data (if this attribute makes sense). By default, when this is set ot ``None``, this rate is recovered from the ``sample_rate`` attribute of the :ref:`Dataset` object, if available. If the latter is not available a ``ValueError`` is raised, and the ``fs`` parameter must be specified
    return_type : str, optional
        This parameter determines the output of this method, and can take the values ``'dataset'`` or ``'array'``; a ``ValueError`` is rised if it doesn't match any of the latter. If this is set to ``'dataset'`` (default), a new :ref:`Dataset` object is created next to the current one, with the ``'_psd'`` string attached to its name, containing the data corresponding to frequencies and PSD of computed with this method. If ``return_type`` is set to ``'array'``, the usual output of `scipy.signal.welch <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_ is returned instead
    dts_key : str, optional
        If ``return_type=='dataset'`` , this is the *key* associated to this dataset. If not set, custom name is chosen, which is the name of the current dataset followed by ``_psd``. Notice that attempting to call this method twice without specifying the name will rase: ``RuntimeError: Unable to create link (name already exists)``
    **psdkwargs : dict, optional
        Any other optional keyword argument accepted by `scipy.signal.welch <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_, with the exception of ``nperseg`` (replaced by ``fftlength``), ``noverlap`` (replaced by ``overlap``), ``nfft`` (automatically set to the next power or two of ``nperseg``), and ``axis`` (``-1``, only 1D arrays). Available arguments are: ``window`` (default ``blackman``), ``detrend`` (``'constant'``), ``return_onesided`` (``True``), ``scaling`` (``'density'``), and ``average`` (``'median'``)
            
    Returns
    -------
        dataset
            If ``return_type`` is ``'dataset'``, this method will ruturn a :ref:`Dataset` object that can be associated to a variable
        f : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
            Array of sample frequencies, with `Astrpy.units <https://docs.astropy.org/en/stable/units/>`_, if ``return_type`` is ``'array'``
        Pxx : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
            Power spectral density or power spectrum of the time series in the dataset, with `Astrpy.units <https://docs.astropy.org/en/stable/units/>`_, if ``return_type`` is ``'array'``
    
    Raises
    ------
    ValueError
        When the data is not numeric or 1D, the sampling frequency is not recovered from the data, or when the ``return_type`` attribute is set to something different than ``'array'`` or ``'dataset'``
    RuntimeError
        When this method is called twice with ``return_type=='dataset'`` and different ``dts_key`` parameters are not specified
    
    Notes
    -----
    Refer to `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated /scipy.signal.get_window.html#scipy.signal.get_window>`_ for a list of the available window functions to pass to the PSD estimation.
    
    If you select ``'dataset'`` as the output type, the resulting data will be in the form of a `structured array <https://numpy.org/doc/stable/user/basics.rec.html#module-numpy.doc.structured_arrays>`_ with ``freq`` and ``PSD`` as the *names* of its two *fields*.
        
    """
    if return_type not in ('dataset','array'):
        raise ValueError("Unrecognised 'return_type' parameter. Values must be either 'dataset' or 'array'.")
    
    if (self[...].ndim != 1):
        raise ValueError("This dataset appears to be not 1D. Unclear meaning to be attributre to the PSD of a multidimensional sequence.")
    else:
        pass
    
    if fs is None:
        try:
            fs = int(self.attrs['sample_rate'])
        except:
            raise ValueError("Unrecognised sampling frequency of the dataset. Please provide one to the 'fs' parameter.")
            
    from scipy.signal import welch
    
    # fftlength
    if fftlength is None:
        nperseg = len(self.data) 
    else:
        nperseg = int(fftlength*fs)
    
    # noverlap
    if overlap is None:
        noverlap = 0
    else:
        noverlap = int(overlap*fs)
    
    nfft=psdkwgs.get('nfft', 2**(nperseg - 1).bit_length())
    
    freq, Pxx = welch(self[...], fs=fs, window=psdkwgs.get('window','blackman'), nperseg=nperseg, noverlap=noverlap, nfft=nfft,
                      detrend=psdkwgs.get('detrend','constant'), return_onesided=psdkwgs.get('return_onesided',True),
                      scaling=psdkwgs.get('scaling','density'), axis =- 1, average=psdkwgs.get('average','median'))
    
    if return_type == 'array':
        # The output array has units
        return freq *apu.Hz, Pxx * (self.attrs.get('unit', '1')/apu.Hz)
    
    elif return_type == 'dataset':  
        # Define a structured data type
        struct_data = np.array(list(zip(freq,Pxx)), dtype=([('freq',np.float),('PSD',np.float)]))
        
        # Get the name of the parent group to save the PSD along with it
        if dts_key is None:
            string_to_append = '_psd'
            dts_key = self.name+string_to_append
            grp = self.parent
        else:
            grp = self.file
        psd_dset = grp.create_dataset(dts_key, data=struct_data)
        
        # Add the attributes
        psd_dset.attrs['f_nyquist'] = fs/2
        
        if 'channel' in self.attrs:
            psd_dset.attrs['channel'] = self.attrs['channel']
        try:
             psd_dset.attrs['unit'] = self.attrs['unit']+"/Hz"  
        except:
            psd_dset.attrs['unit'] = "1/Hz"
        
        return psd_dset

@_add_method(h5py.Dataset)
def resample(self, outfs, fs=None, method='poly', return_type='dataset', dts_key=None, fraclimit=None, **reskwgs):
    """
    Resample ``self`` along the given axis using various possible methods, specified by the ``method`` argument. Each of them has some advantages and disadvantages, so a careful choise is advisable.
    
    ``poly`` (default) is based on `scipy.signal.resample_poly <https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.signal.resample_poly.html#scipy.signal.resample_poly>`_ does upsampling/downsampling and the associated filtering, using a `polyphase filter <https://en.wikipedia.org/wiki/Polyphase_quadrature_filter>`_. This has a very good balance in speed and performance, with only a moderate distortion of the data close to the Nyquist frequency. If ``outfs`` and ``fs`` are prime, this can be quite slow too.
    
    ``fft`` is based on `scipy.signal.resample method <https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.signal.resample.html>`_. It does upsampling/downsampling and the associated filtering, entirely in the frequency domain, using the `Fast Fourier Transform technique <https://en.wikipedia.org/wiki/Trigonometric_interpolation#Relation_with_the_discrete_Fourier_transform>`_ (bonus for data length that are multiple of two). Because it is using the Fourier Transform, a key assumption is that the signal that is fed in its input is *periodic*. This gives usually the best performances and mantaines the spectrum unaltered. However it can be slow if the input data lenght is not multiple of two and prime, and also if it is not multiple of the output samples. In the latter case, a slight distorsion of the time axis is also possible, with a linearly increasing phase difference. Avoid this method if this is the case.
    
    ``decimate`` is arguably the simplest and fastest method (``ftype='iir'``), although it is not always advisable to use it and the overal performances are not as good as the previous two methods. It is based on the `scipy.signal.decimate method <https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.signal.decimate.html>`_, which consists in an antilaiasing low-pass filter (either an ``'iir'`` or a ``'fir'`` filter) followed by a decimation, such as ``data[::q_factor]``. If a ``'iir'`` filter is chosen and the relative *q_factor* between input and output frequency is larger than 10, the decimate method is called recursively. For ``iir``, the distorsion can be significant in proximity of the Nyquist frequency. Selecting ``'fir'`` as ``ftype`` we will have a linear phase, the output will be the same of ``poly``, but the method will result generally slower. If we select ``'iir'``, it is going to be much faster, but the phase is only approximately linear. Also, differently from the other two methods, this is not suitable for *upsampling* or when the original sampling frequency is not a multiple of ``outfs``. In this case ``poly`` is used instead.
   
    Parameters
    ----------
    outfs : float
        Desired output frequency of the resampled data. It is advisable to pick a frequency that is submultiple of the original sampling frequency of the data
    fs : float, optional
        Sampling frequency of ``self``. If not provided, this is automatically recovered from ``self.attrs['sample_rate']``. If this attribute is not present an ``AttributeError`` is raised.
    return_trype : str, optional
        array, dataset, inplace
    method : str, optional
        Method with which to perform the resampling. Default is ``'poly'``, for a *poly-phase filtering*. Don't modify it if you don't have read the description above first
    return_type : str, optional
        This parameter determines the output of this method, and can take the values ``'dataset'`` or ``'array'``; a ``ValueError`` is rised if it doesn't match any of the latter. If this is set to ``'dataset'`` (default), a new :ref:`Dataset` object is created next to the current one, with the ``'_r{outfs}'`` string attached to its name, containing the data corresponding to resampled time series. If ``return_type`` is set to ``'array'``, the output is an array  
    dts_key : str, optional
        If ``return_type=='dataset'`` , this is the *key* associated to this dataset. If not set, custom name is chosen, which is the name of the current dataset followed by ``f'_r{outfs}''``. Notice that attempting to call this method twice without specifying the name will rase: ``RuntimeError: Unable to create link (name already exists)``
    fraclimit : int, optional
        What if one attempts to reample to ``outfs = np.pi``? To prevent this questionable choice, still allowing float values for the input and output frequencies, we have introduced this parameter as the maximum denominator of the fraction equivalent to ``fs``/``outfs``. This also prevent rounding issues when working with floats. Default value is ``1000``. Try to make use of integer frequencies, instead of changing this value  

    **reskwgs : dict, optional
        These are all the other keyword arguments to pass to the previous resampling method. Refer to the corresponding documentation for further details
    Returns
    -------
    dataset
        If ``return_type`` is ``'dataset'``
    array
        If ``return_type`` is ``'array'``
    
    Raises
    ------
    AttributeError
        If ``fs`` is not specified and a valid ``'sample_rate'`` attribute is not present
    ValueError
        If ``method`` aor ``return_type`` take values that are not permitted
    RuntimeError
        When this method is called twice with ``return_type=='dataset'`` and different ``dts_key`` parameters are not specified
    
    See Also
    --------
    gwdama.preprocessing.decimate_recursive
        
    """
    from fractions import Fraction

    if not fs:
        fs = int(self.attrs['sample_rate'])
        
    if method=='poly':
        from scipy.signal import resample_poly
        
        # Allow float frequencies
        if not isinstance(fs,int):
            fs=Fraction(fs)
            
        if not isinstance(outfs,int):
            outfs=Fraction(outfs)

        Frac = Fraction(fs, outfs).limit_denominator(fraclimit or 1000)
        up, down = Frac.denominator, Frac.numerator
        res = resample_poly(self[...], up=up, down=down, **reskwgs)
    elif method=='fft':
        from scipy.signal import resample

        ax = reskwgs.get('axis',0)
        num = int(self.shape[ax] * outfs/fs)
        res = resample(self[...], num=num, **reskwgs)
    elif method=="decimate":
        from gwdama.preprocessing import decimate_recursive
        from scipy.signal import decimate

        # Allow float frequencies
        if not isinstance(fs,int):
            fs=Fraction(fs)
            
        if not isinstance(outfs,int):
            outfs=Fraction(outfs)

        Frac = Fraction(fs, outfs).limit_denominator(fraclimit or 1000)
        up, down = Frac.denominator, Frac.numerator
        if up==1:
            ftype = reskwgs.get('ftype','fir')
            if ftype=='fir':
                reskwgs.update({"ftype": "fir"})
                res = decimate(self[...], q=down, **reskwgs)
            else:
                res = decimate_recursive(self[...], q_factor=down, **reskwgs)[0]
        else:
            print("Warning!! The output frequency is not a submultiple of the original sampling. Decimation mathod not compatible, 'poly_phase' filter applied instead")
            from scipy.signal import resample_poly
            res = resample_poly(self[...], up=up, down=down)
    else:
        raise ValueError("Unrecognised 'method' parameter. Values must be either 'poly' (poly-phase filter), 'fft', or 'decimate' (lowpass+decimation).")
        
    if return_type=='array':
        return res
    
    elif return_type=='dataset':
        if dts_key is None:
            string_to_append = f'_r{outfs}'
            dts_key = self.name+string_to_append
            grp = self.parent
        else:
            grp = self.file            
        res_dset = grp.create_dataset(dts_key, data=res)
        for k in self.attrs.keys():
            res_dset.attrs[k] = self.attrs[k]
        if isinstance(outfs,Fraction):
            res_dset.attrs['sample_rate'] = float(outfs)
        else:
            res_dset.attrs['sample_rate'] = outfs
        return res_dset        
        
    else:
        raise ValueError("Unrecognised 'return_type' parameter. Values must be either 'dataset' or 'array'.")
        
    
@_add_method(h5py.Dataset)   
def taper(self, fs=None, side='leftright', duration=None, nsamples=None, return_type='dataset', dts_key=None, window=('tukey',0.25)):
    """
    Taper the edges of this datset smoothly to zero. The method automatically tapers from the second stationary point (local maximum or minimum) on the specified side of the input. However, the method will never taper more than half the full width of the data, and will fail if there are no stationary points.
    
    Parameters
    ----------
    fs : float, optional
        Sampling frequency of this dataset. By defoult (``fs=None``) this value is read from the ``sample_rate`` attribute of the dataset, if it exists. If not, this is automatically assumed to be 1
    side : str, optional
        Sides to smooth to zero. Possible options are: ``leftright`` (defautl), ``left``, or ``right``
    duration : float, optional
        The duration of time to taper, will override ``nsamples`` if both are provided as arguments
    nsamples : int, optional
        The number of samples to taper, will be overridden by ``duration`` if both are provided as arguments
    return_type: str, optional
        This parameter determines the output of this method, and can take the values ``'dataset'`` or ``'array'``; a ``ValueError`` is rised if it doesn't match any of the latter. If this is set to ``'dataset'`` (default), a new :ref:`Dataset` object is created next to the current one, with the ``'_taper'`` string attached to its name, containing the data corresponding to the whitened timeseries
    dts_key : str, optional
        If ``return_type=='dataset'`` , this is the *key* associated to this dataset. If not set, custom name is chosen, which is the name of the current dataset followed by ``_taper``. Notice that attempting to call this method twice without specifying the name will rase: ``RuntimeError: Unable to create link (name already exists)``
    
    Returns
    -------
    dataset
        If ``return_type`` is ``'dataset'``, this method will ruturn a :ref:`Dataset` object that can be associated to a variable
    whitened : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ 
        Whitened timeseries
        
    Raises
    ------
    RuntimeError
        When this method is called twice with ``return_type=='dataset'`` and different ``dts_key`` parameters are not specified   
    
    """
    from gwdama.preprocessing import taper
    
    if return_type not in ('dataset','array'):
        raise ValueError("Unrecognised 'return_type' parameter. Values must be either 'dataset' or 'array'.")
    
    if (self.data.ndim != 1):
        raise ValueError("This dataset appears to be not 1D. Plase, provide 1D sequence.")
    else:
        pass
    
    if not fs:
        try:
            fs = int(self.attrs['sample_rate'])
        except:
            fs = 1
            pass
    
    tp_data = taper(self[...], fs=fs, side=side, duration=duration, nsamples=nsamples, window=window)

    if return_type == 'array':
        return tp_data
    
    elif return_type == 'dataset':
        if dts_key is None:
            string_to_append = '_taper'
            dts_key = self.name+string_to_append
            grp = self.parent
        else:
            grp = self.file
        tp_dset = grp.create_dataset(dts_key, data=tp_data)
        
        tp_dset.attrs['taper'] = window
        for k in self.attrs.keys():
            tp_dset.attrs[k] = self.attrs[k] 
        
        return tp_dset
    

@_add_method(h5py.Dataset)
def to_TimeSeries(self, fs=None, unit=None, t0=None, **TSkwgs):
    """
    This method attemp to convert the current :class:`~gwdama.io.Dataset` into a `TimeSeries instance of GWpy <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_. The array-like ``value`` is taken from ``self.data``. The other parameters are either read from the :class:`~gwdama.io.Dataset` attributes or set manually passing them as arguments to the method.
    
    Parameters
    ----------
    fs : float, optional
        The sampling rate of the current dataset. If not specified, this method attempts by default to read the value of the ``sample_rate`` attribute, if available. If not, ``fs`` is set to 1. Notice that this corresponds to the ``sample_rate`` parameter of the `TimeSeries class <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_
    t0 :  LIGOTimeGPS, float, str, optional
        GPS epoch associated with these data, any input parsable by to_gps is fine
    unit : `Unit <https://docs.astropy.org/en/stable/api/astropy.units.Unit.html#astropy.units.Unit>`_, optional
        physical unit of these data
    **TSkwgs : dict, optional
        These are all the other keyword arguments to pass to the `TimeSeries class <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_. Refer to the corresponding documentation for further details
        
    Returns
    -------
    : `TimeSeries <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_
        The current dataset converted into an `instance of GWpy TimeSeries <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_
    
    """
    if not fs:
        fs = self.attrs.get('sample_rate', 1)
    if not t0:
        t0 = self.attrs.get('t0', 0)
    if not unit:
        unit = self.attrs.get('unit', None)
        
    TS = TimeSeries(self[...], t0=t0, sample_rate=fs, unit=unit, **TSkwgs)
    return TS
    
    
@_add_method(h5py.Dataset)
def whiten(self, fftlength=None, overlap=None, fs=None, phase_shift=0, time_shift=0, return_type='dataset', dts_key=None, taper_edges=None, **psdkwgs):
    """
    Method to compute the *whiten time-series* from the current dataset: the input is detrended and the output normalised such that, if the input is stationary and Gaussian, then the output will have zero mean and unit variance. The most of the parameters of this mehtod are those needed to compute a PSD. A tapering of the edges is also applied to get rid of the effect of filter settle-in.
    
    Parameters
    ----------
    fftlength : float, optional
        Length in seconds of the segment to use for computuing the psd
    taper_edges : float, optional
        The edges of the whitened time-series can be round off to reduce filtering effects. You can specify the duration of the taper with this parameter or set it to 0 if you don't want any. Dafault duration is half of ``fftlength``
    dts_key : str, optional
        If ``return_type=='dataset'`` , this is the *key* associated to this dataset. If not set, custom name is chosen, which is the name of the current dataset followed by ``_whiten``. Notice that attempting to call this method twice without specifying the name will rase: ``RuntimeError: Unable to create link (name already exists)``
    
    Raises
    ------
    RuntimeError
        When this method is called twice with ``return_type=='dataset'`` and different ``dts_key`` parameters are not specified
   
    See Also
    --------
    gwdama.io.dataset.psd : Reference for the parameters to be chosen in this method
    gwdama.io.dataset.taper : Tapering edges technique
            
    """
    from scipy.interpolate import interp1d
    from gwdama.preprocessing import whiten, taper
    
    if not fs:
        try:
            fs = int(self.attrs['sample_rate'])
        except:
            raise ValueError("Unrecognised sampling frequency of the dataset. Please provide one to the 'fs' parameter.")
    
    if not fftlength:
        fftlength = min(4, self.duration())
    if not overlap:
        overlap = fftlength/2
    
    pkwgs = {'detrend':'linear', 'return_onesided':True, 'scaling':'density'}
    pkwgs.update(psdkwgs)
    # Get the PSD as an array
    freq, PSD = self.psd(fftlength=fftlength, overlap=overlap, fs=fs, return_type='array', **pkwgs)
    
    interp_psd = interp1d(freq, PSD, fill_value="extrapolate")
    
    # Detrend the data before applying the whitening function
    from scipy.signal import detrend
    ddata = detrend(self[...])
    white = whiten(ddata, interp_psd, 1/fs, phase_shift=phase_shift, time_shift=phase_shift)
    
    # Due to filter settle-in, a segment of length 0.5*fduration will be corrupted at the beginning and end
    # of the output. 
    if psdkwgs:
        win = psdkwgs.get('window', 'tukey')
    else:
        win = ('tukey',0.5)
    
    if taper_edges is None:
        white = taper(white, fs=fs, duration=fftlength/2, window=win)
    elif taper_edges>0:
        white = taper(white, fs=fs, duration=taper_edges, window=win)
    else:
        pass
    
    if return_type == 'array':
        return white
    
    elif return_type == 'dataset':
        if dts_key is None:
            string_to_append = '_whiten'
            dts_key = self.name+string_to_append
            grp = self.parent
        else:
            grp = self.file
        white_dset = grp.create_dataset(dts_key, data = white)
        
        for k in self.attrs.keys():
            white_dset.attrs[k] = self.attrs[k] 
        
        return white_dset
    
@_add_property(h5py.Dataset)
def show_attrs(self):
    """
    Method that make a Dataset to 'show the attributes'
    """
    to_print = ''
    for k, val in self.attrs.items():
        to_print += "{:>10} : {}\n".format(k, val)
    #return to_print
    print(to_print)

# ----- Experimental -----
@_add_method(Figure)
def reshow(self):
    """
    Method to re-show a closed Figure object
    """
    from gwdama.plot import reshow
    reshow(self)
    


"""
This submodule contains some function to produce nice plots 'out of the box'
"""

import matplotlib.pyplot as plt

def make_hist(data, xlabel='Value', figsize=(10,6), title=None, ax=None, **kwargs):
    """
    Function to prepare plots.
    """
    
    default_kwgs = {
        'color':'#78909c',
        'bins':'auto',
        'alpha': .6,
        'rwidth': 0.9,
        'density': True
        }
        
    for k, v in default_kwgs.items():
        kwargs[k] = kwargs.get(k, v)
    
    # If an axis object already exists, use it
    if ax:
        axh = ax
    else:
        fig, axh = plt.subplots(figsize=figsize)
        
    # An "interface" to matplotlib.axes.Axes.hist() method
    n, bins, patches = axh.hist(x=data, **kwargs)
    axh.grid(alpha=0.4)
    axh.set_xlabel(xlabel)
    axh.set_title(title)
    if kwargs['density']:
        axh.set_ylabel('Density')
    else:
        axh.set_ylabel('Occurrence')

    if ax:
        return axh
    else:
        return fig
    
def reshow(fig):
    """This function creates a dummy figure and uses its manager to
    display ``fig``. This is useful when you want to show again a
    pyplot Figure object that has been closed.
    """
    plt.close(fig)
    dummy = plt.figure()
    new_manager = dummy.canvas.manager
    new_manager.canvas.figure = fig
    fig.set_canvas(new_manager.canvas)


def plot_itf_meta(meta, t0=None, sample_rate=None, mode='bar'):
    """
    WORK in Progress
    Function to plot 'V1:DQ_META_ITF_Mode'
    For rference:
    ITF_mode: 	V1:DQ_META_ITF_Mode:
             1: Science
             0: Unknown
            -1: Adjusting
            -3: Commissioning
            -4: Maintenance
            -5: Calibration
            -7: Locked
            -8: Not locked
            -9: Locking
            -11:Upgrading
            -12:Troubleshooting
    """
    
    
    labels = ['Science', 'Unknown', 'Adjusting', 'Commissioning',
              'Maintenance','Calibration','Locked','Not locked',
              'Locking','Upgrading','Troubleshooting']

    colors = ['green','red','fuchsia', 'darkorange',
              'firebrick', 'darkviolet','blue', 'gold',
              'lightskyblue', 'lightcoral','crimson']

    #metadict = {'Science': }
    
    ax1.set_title("S200311ba - "+channel)
    ax2.axvspan(ti[0],tl, color='g', label='Science {:.2f}%'.format(100 * sum(meta.data==1)/meta.size))
    ax2.axvspan(tl,tr, color='magenta', label='Adjusting {:.2f}%'.format(100 * sum(meta.data==-1)/meta.size))
    ax2.axvspan(tr,ti[-1], color='g')
    ax2.set_xlabel(f'Time [sec] around {from_gps(e_gps)} UTC ({e_gps})')
    ax2.set_xlim(ti[0],ti[-1])
    ax2.set_yticks([])
    leg = ax2.legend(title='V1:DQ_META_ITF_Mode', bbox_to_anchor=(1.02, 1.5), loc='upper left',)
    leg._legend_box.align = "left"
    
    
    return fig


"""
Utility functions
"""
# To define wrappers to enhance existing classes
from functools import wraps
from gwosc.datasets import find_datasets
from gwosc.datasets import event_gps, run_segment 
from gwpy.time import to_gps

def _add_property(cls):
    """
    Decorator to add properties to already existing classes.
    Partially based on the work of M. Garod:
    https://medium.com/@mgarod/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6
    
    Example
    -------
    To add theproperty `func` to the class Class:
    @_add_property(Class)
    def func(self):
        pass
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs): 
            return func(self,*args, **kwargs)
        setattr(cls, func.__name__, property(wrapper))
        # Note we are not binding func, but wrapper which accepts self but does exactly the same as func
        return func # returning func means func can still be used normally
    return decorator


def _add_method(cls):
    """
    Decorator to add methds to already existing classes.
    Partially based on the work of M. Garod:
    https://medium.com/@mgarod/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6
    
    Example
    -------
    To add theproperty `func` to the class Class:
    @_add_method(Class)
    def func(self):
        pass
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs): 
            return func(self,*args, **kwargs)
        setattr(cls, func.__name__, wrapper)
        # Note we are not binding func, but wrapper which accepts self but does exactly the same as func
        return func # returning func means func can still be used normally
    return decorator

# def name_changer(obj, kay):
#     if not os.path.exists(outdir):
#         os.makedirs(outdir)
#     else:
#         expand = 0
#         while True:
#             expand += 1
#             new_event = event_id +'_'+ str(expand)
#             if os.path.exists(args.directory.rstrip('/')+'/'+new_event):
#                 continue
#             else:
#                 event_id = new_event
#                 break
#         outdir = args.directory.rstrip('/')+'/'+event_id
#         os.makedirs(outdir)	

def find_run(start, end, host='https://www.gw-openscience.org'):
    """
    Given GPS ``start`` and GPS ``stop``, it returns the run the data belongs to,
    otherwise it rises an error: data not belonging to gwosc
    
    Parameters
    ----------
    start : LIGOTimeGPS, float, str
        GPS start time of required data; any input parseable by to_gps is fine
    end : LIGOTimeGPS, float, str
        GPS end time of required data; any input parseable by to_gps is fine
    host : str, optional
        The URL of the GWOSC host to query, defaults to ``'https://www.gw-openscience.org'``
    
    """
    if type(start) != type(to_gps('now')):
        start = to_gps(start)
    if type(end) != type(to_gps('now')):
        end = to_gps(end)

    # O3a
    start_a = to_gps('2019-04-01')
    end_a = to_gps('2019-10-02')
    # O3b
    start_b = to_gps('2019-10-31')
    end_b = to_gps('2020-03-28')
    
    if (start>=start_a) and (end<=end_a):
        run = 'O3A'
    elif (start>=start_b) and (end<=end_b):
        run = 'O3B'        
    
    else:
        all_runs = find_datasets(type='run',segment=(start.gpsSeconds, end.gpsSeconds),host=host)
        
        if all_runs:
            # Remove this not to mess up with things
            try:
                all_runs.remove('BKGW170608_16KHZ_R1')
            except:
                pass

            run = all_runs[0]
            for r in all_runs:
                if r[:2] in run:
                    pass
                else:
                    raise Exception('Data spanning multiple runs! Try to reduce the gps time interval.') 
            run = run[:2]
        else:
            raise Exception('GPS time interval not corresponding to any known run. Please, double-check ``start`` and ``stop``.')
    return run
    


# This file is part of the GwDataManager package.
# Import data from various location into a dictionary-like object.
# 

"""
GwDataManager
=============

This is the basic container to read and store data, and give acces to their manipulation in term of Datasets

This is based on `h5py` classe File and Datase, with some modifications.
In particular, the class Dataset comprizes some new methods and properties
to make it more easy to manipulate.
"""

import sys, time
import numpy as np
from os.path import join, isfile, split, splitext, basename, dirname
from glob import glob
from multiprocessing import Pool, cpu_count

# For the base classes and their extentions
import h5py
from .dataset import * 
from gwdama.utilities import find_run, _add_method

# For File object stored in memory or as temprary file
from io import BytesIO
from tempfile import TemporaryFile

# Locations of ffl
from . import ffl_paths

# imports related to gwdama
from .gwLogger import GWLogger

# Necessary to fetch open data and to handle gps times
from gwpy.timeseries import TimeSeries, TimeSeriesDict
from gwpy.time import to_gps

# Necessary to export or import as pandas Dataframe
from pandas import DataFrame

# To print warnings
from warnings import warn

# ----- Utility functions specific to GwDataManager -----

def recursive_struct(group, indent = '  '):
    """
    Function to print a nice tree structure:
    │
    ├──
    └──
    """
    to_print = ''
    if isinstance(group, h5py.Group):
        for i,k in enumerate(group.keys()):
            if i+1<len(group.keys()):
                to_print += indent + f"├── {k}\n"
            else:
                to_print += indent + f"└── {k}\n"
                
            if isinstance(group[k], h5py.Group) and (i+1<len(group.keys())):
                to_print += recursive_struct(group[k], indent=indent+'│   ')
            elif isinstance(group[k], h5py.Group):
                to_print += recursive_struct(group[k], indent=indent+'    ')
                  
    return to_print 

def key_changer(obj, key=None):
    """
    Check if a key is present in a dictionary-like object.
    If already present, change its name adding '_#', with increasing
    numbers.
    """
    if key is None:
        key = 'key'  # Default value
    if key in obj.keys():
        expand = 0
        while True:
            expand += 1
            new_key = key +'_'+ str(expand)
            if new_key in obj.keys():
                continue
            else:
                key = new_key
                break
    return key

# ----- Additional method of h5[y.Group to return Datasets -----

@_add_method(h5py.Group)
def to_DataFrame(self, **dfkwgs):
    """
    Convert this Group to a `pandas.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_. The names of the columns are the keys of each Dataset, and their numeric value is equal to their data. The Datasets must have the same length in order to be aligned, otherwise a ``TypeError`` will be raised.
    
    Parameters
    ----------
    dfkwgs : dict, optional
        Keyword arguments of `pandas.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_, excluding ``data``. By default, the ``index`` parameter is obtained from the ``times`` attribute of each Dataset if they are equal. Otherwise it is a RangeIndex
        
    Returns
    -------
     : `pandas.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
         DataFrame representing the current `Group <https://docs.h5py.org/en/stable/high/group.html>`_ object
     
    Raises
    ------
    TypeError
        if the Datasets in this group have different lengths and can't be aligned into a dataframe
    """
    # Check that the length, t0 and sample rate of all the elemts of the group are the same
    chks = {'len': [], 't0': [], 'sample_rate': []}
    for val in self.values():
        if isinstance(val, h5py.Dataset):
            chks['len'].append(len(val))
            chks['t0'].append(val.attrs.get('t0', np.nan))
            chks['sample_rate'].append(val.attrs.get('sample_rate', np.nan))

    if not all(ll==chks['len'][0] for ll in chks['len']):
        raise TypeError("This group has Datasets of different lengths. Cannot align them to a Pandas DataFrame. Maybe this is due to different 'sample_rate' and 'duration' attributes. Try resampling, in this case")
    elif all(tt==chks['t0'][0] for tt in chks['t0']) and all(sr==chks['sample_rate'][0] for sr in chks['sample_rate']):            
        index = np.arange(chks['len'][0])/chks['sample_rate'][0] + chks['t0'][0]
        index_name = 'gps_time'
    else:
        warn("WARNING! The Datasets have not the same initial time or sample rate. The index of the resulting DataFrame is simply a RangeIndex")
        index = np.arange(chks['len'][0])
        index_name = None
    
    dfkwgs['index'] = dfkwgs.get('index', index)
    
    df = DataFrame(data={k: val[...] for k, val in self.items() if isinstance(val, h5py.Dataset)},**dfkwgs) 
    df.index.name = index_name
    
    return df
    
# ----- The Class -----

class GwDataManager(h5py.File):
    """
    Main class for GW data management. Basically, this is an enhanced version of the :py:class:`h5py.File` class, with various methods for importing GW data from different sources, pre-processing functions, plotting, and more.
    
    One key aspect regards the ``storage`` of this file while in use, as specified by the corresponding parameter. This can be an actual *hdf5 file* saved on ``'disk'`` (to be passed to the previous parameter), or a `file-like object <https://docs.h5py.org/en/stable/high/file.html#python-file-like-objects>`_, such as a temporary file (``'tmp'``) on disk, or a `BytesIO file <https://docs.python.org/3/library/io.html#io.BytesIO>`_ using a in-memory bytes buffer (``'mem'``). The first method is the more common when manipulating existing files. The latter ``io.BytesIO``, being in-memory, is usually the fastest and particularly indicated for testing. If you want to write large amounts of data, a better option may be to store temporary data on disk using ``'tmp'``.
    
    Also, the parameter ``mode`` can be passed to the class instance. If ``storage='disk'``, this tells how to open the file; possible vallues are: ``'r'``, ``'r+'``, ``'a'``, ``'w'``, ```'w-'`` or ``'x'``. If the previous option allow modifying the file, then new memory is allocated on disk next to the one occupied by the source file. It is important to close the file once you are done with manipulating it, ``dama.close()``, or to make use of a `context manager <https://book.pythontips.com/en/latest/context_managers.html>`_: ``with GwDataManager('/path/to/file.h5', mode='r+') as dama: ...``.
    
    Instead, if the :class:`~gwdama.io.GwDataManager` istance refers to a file-like object (``'tmp'`` or ``'mem'``), one can can pass ``mode='r'``, and import the data stored in the hdf5 file at the path specified by ``dama_name``. In this case, a copy of the provided file is loaded in-memory or as a temporary file. Remeber to save it afterwards. An ``IOError`` is raised if this is not a valid path to a file. If ``mode`` is not specified (or if it has any other value), a blank file-like object is created.
    
    Parameters
    ----------
    dama_name : str, optional
        Name to assign to the :class:`~gwdama.io.GwDataManager` object or path to an existing hdf5 file if the ``mode`` parameter is set to this purpose (that is, ``'r'``, ``'r+'`` or ``'a'``). A defoult ``'mydama'`` is chosen if not provided.
    storage : str, optional
        This parameter determines where the bytes corresponding to this object are stored. Available options are ``'disk'``, useful when one wants to modify an existing "dama" (hdf5), or ``'mem'`` and ``'tmp'`` for a "file-like" object. In the case of ``'mem'`` (default), a BytesIO file is created in-memory. This is the fastest option and particularly recommended for testing. `If `'tmp'``, a temporary file is created on disk (but it is not accessible otherwise). This is the preferreed option when manipulating very large amounts of data.
    mode : str, optional
        When the ``storage`` parameter is set to ``'disk'``, this parameter determines the mode the file specified in ``dama_name`` is opened. Avaliable options are: ``'r'`` (read only, file must exist), ``'r+'`` (read/write, file must exist), ``'w'`` (create file, truncate if exists), ``'w-'`` or ``'x'`` (create file, fail if exists), and ``'a'`` (read/write if exists, create otherwise). The default behaviour is that ``mode='r'`` if ``storage='disk'``, that is, it attempts to read an existing file on disk if this option is chosen. Raises error if the file is not present. Instead, for the other available options of ``storage``, this parameter has no effect but in the case ``dama_name`` points to an existing file and this ``mode`` is set to ``'r'``. In this case, a new file like object is created with the content of the existing file (which is not modified otherwise).
    **kwargs : dict, optional
        Other parameters that can be passed to `h5py.File objects <https://docs.h5py.org/en/stable/high/file.html>`_. 
    
    Raises
    ------
    OSError
        if ``mode`` is ``'w-'`` or ``'x'`` and the file you want to create already exists, or with ``'r+'`` or ``'r'`` if the file you want to read doesn't exist.
    
    Examples
    --------
    Everything starts by initializing the :class:`~gwdama.io.GwDataManager` class like this::
    
      >>> dama = GwDataManager()
      
    If you want to assign a name to this instance, pass it within parentesis. By default, this is an in-memory object that will be cancelled when you close the Python session. However, it is good practice to close it beforehand with::
    
      >>> dama.close()
      
    Instead, if you want to open an existing hdf5 file, such as a previously created "dama", pass its name as the ``dama_name``, and make use of the value ``'r'`` (or similar) for the parameter ``mode``.
        
    Notes
    -----
    Refer to the description for more details on the parameter ``storage``, which determines the memory allocation of this class.
         
    """
    # NOTE: put methods in alphbetical order!
   
    def __init__(self, dama_name="mydama", storage='mem', mode=None, **kwargs):
        
        # Create a new temporary file. If dama_name is a path to a valid file, its content is recovered
        if storage in ('mem','tmp'):
            if storage=='mem':
                tf = BytesIO()
            elif storage=='tmp':
                tf = TemporaryFile()
            # Create in-memory object  
            super().__init__(tf, mode='w', **kwargs)    
            
            # Add a couple of attributes to this dataset, like a name and its time stamp
            self.attrs['dama_name'] = splitext(split(dama_name)[-1])[0]
            self.attrs['time_stamp']=str(time.strftime("%y-%m-%d_%Hh%Mm%Ss", time.localtime()))
            
            # A file exists at the path 'dama_name', try to get its content
            if isfile(dama_name) and (mode=='r'):
                print("Reading dama... ",end='')
                with h5py.File(dama_name, 'r') as existing_dama:
                    for k in existing_dama.keys():
                        existing_dama.copy(k,self)
                    for k in existing_dama.attrs.keys():
                        self.attrs[k] = existing_dama.attrs[k]
                    print(' done.')
                            
        elif storage=='disk':
            if isfile(dama_name) and (mode is None):   
                print("Reading dama... ",end='')
                super().__init__(dama_name, mode='r', **kwargs)
                print(' done.')

            elif isfile(dama_name):
                print("Reading dama... ",end='')
                super().__init__(dama_name, mode=mode, **kwargs)
                print(' done.')                
                          
            else:
                print('Creating new dama... ',end='')
                super().__init__(dama_name, mode=(mode or 'w-'), **kwargs)
                self.attrs['dama_name'] = splitext(split(dama_name)[-1])[0]
                self.attrs['time_stamp']=str(time.strftime("%y-%m-%d_%Hh%Mm%Ss", time.localtime()))                
                print(' done.')                   

                
    def __repr__(self):
        """
        String representation of the object.
        """
        str_to_print = f"<{type(self).__name__} object at {hex(id(self))}: {self.attrs['dama_name']}>"
        #for k, val in self.__data_dict.items():
        #    str_to_print +="\n  {:>8} : {}".format(k,val.__repr__())
        return str_to_print
 

    def __str__(self):
        str_to_print = f"{self.attrs['dama_name']}:\n"+ recursive_struct(self)
 
        str_to_print += "\n  Attributes:\n"
        for k, val in self.attrs.items():
            str_to_print += "  {:>12} : {}\n".format(k, val)

        return str_to_print

    
    @property
    def show_attrs(self):
        """
        Property that makes the :class:`~gwdama.io.GwDataManager` object to print in a conveninet way all of its attributes name and key pairs.
        """
        to_print = ''
        for k, val in self.attrs.items():
            to_print += "{:>10} : {}\n".format(k, val)
        #return to_print
        print(to_print)

        
    @property
    def groups(self):
        """
        This property returns a list with the name of each group and subgroup af the current :class:`~gwdama.io.GwDataManager` object.
        This fuction could be possibly improved allowing the possibility to create sublists for each group and subgroup.
        """
        groups = []
        self.visititems(lambda g,_: groups.append(str(g)))
        return groups
    
    def read_gwdata(self, start, end, data_source=None, dts_key=None, duplicate='rename', return_output=False, **kwargs):
        """   
        This is the main function to access data from various locations. Given the value of the ``data_source`` parameter, this method can read form files in the ``local`` file system, provided their paths, or some specifications about them, or from `GWOSC <https://www.gw-openscience.org/about/>`_, either online or from CVMFS. Depending on this parameter or on other passed as ``**kwargs``, this function selecs one from        :func:`~gwdama.io.GwDataManager.read_from_virgo`, :func:`~gwdama.io.GwDataManager.read_gwdata_gwosc_remote` or :func:`~gwdama.io.GwDataManager.read_gwdata_gwosc_cvmfs` methods. The corresponding data will be added to the current dataset.
        
        Parameters
        ----------
        start : LIGOTimeGPS, float, str
            GPS start time of required data. Any input parseable by to_gps is fine. For example, a quick way to provide dates is in the format of ``'YYYY-MM-DD (hh:mm:ss)'``
        end : LIGOTimeGPS, float, str, optional
            GPS stop time of required data; any input parseable by to_gps is fine
        data_source : str, optional
            defines the way data is read, and from where. Possible options: ``gwosc-online`` (default), ``gwosc-cvmfs``, ``local``. If none of the previous options hase been passed, then other parameters passed as keyword arguments are checked. For example, ``ffl_spec``, ``ffl_path`` or ``gwf_path``, which make to select option ``local``. Otherwise the default option is ``geosc-pnline``, meaning that the data are accessed from the GWOSC open data
        dts_key : str, optional
            name of the Dataset or of the Group to append to the :class:`~gwdama.io.GwDataManager`. If not provided and ``channels`` is provided and it is a string (single channel), then its name is used as ``dts_key``. If ``channels`` is a list of channels, then the ``dts_key`` is by defoult ``'channels'``
        duplicate : {``'replace'``, ``'rename'``, ``'raise'``}, optional
            If we try to append a dataset with an already existing key, we have the possibility
            to ``'replace'`` the previous one (deleting its content) or ``'rename'`` the corresponding
            key: "existing_key" -> "exisitng_key_1". Otherwise, with ``'rise'`` an error message is risen. Default ``'rename'``
        return_output : bool, optional
            If ``False`` (default), the new :class:`~gwdama.io.Dataset` is added to the :class:`~gwdama.io.GwDataManager` object and no output is returned. If ``True``, ``self['dts_key']`` is returned, which can be a Dataset or a Group depending on the number of channels that have been read
        **kwargs
            Depending on the ``data_source``, this are the keyword arguments for the corresponding direct method.
            For example, if ``data_source=='local'`` refer to the documentation of the method ``read_from_virgo``.
            If ``data_source=='gwosc-online'``, refer to ``read_gwdata_gwosc_remote``, or if ``data_source=='gwosc-cvmfs'`` to ``read_gwdata_gwosc_cvmfs``
            
        Returns
        -------
         : None, Dataset or Group
             Depending on the ``return_output`` parameter, and the number of channels that have been read
            
        Raises
        ------
        RuntimeError
            every time a wrong data source is provided
        
        Examples
        --------
        To import data from online gwosc (default), we only need to pass a time interval and the name of the interferometer whose data we are interested to: 
        
        >>> from gwdama.io import GwDataManager
        
        >>> start = "2017-08-14 11:59" # One minute before GW170814 merger
        >>> end   = "2017-08-14 12:00:10"
        >>> dama = GwDataManager()  # Default name 'mydama' assigned to the dictionary
        >>> dama.read_gwdata(start, end, ifo='L1')
        
        See Also
        --------
        :func:`~gwdama.io.GwDataManager.read_from_virgo`
        :func:`~gwdama.io.GwDataManager.read_gwdata_gwosc_remote`
        :func:`~gwdama.io.GwDataManager.read_gwdata_gwosc_cvmfs`
        
        """
        # Define a key if not provided:
        # - the name of the channel if there is only one
        # - 'channels' if there are more
        # - strain if we read from gwosc
        if not dts_key:
            if 'channels' in kwargs:
                if isinstance(kwargs['channels'], str):
                    dts_key = kwargs['channels']
                else:
                    dts_key = 'channels' # In this case this is a group
            else:
                dts_key = 'strain'
        
        # Define the data source if not provided
        if not data_source:
            if ('ffl_path' in kwargs) or ('ffl_spec' in kwargs) or ('gwf_path' in kwargs):
                data_source = 'local'
            else:
                data_source = 'gwosc-online'
        
        if (duplicate == "replace") and (dts_key in self.keys()):
            del self[dts_key]
        elif (duplicate == "rename") and (dts_key in self.keys()):
            dts_key = key_changer(self, key=dts_key)
        elif (dts_key in self.keys()):
            raise Exception("Unrecognised 'duplicate' parameter. Please, select either 'rename' or 'replace'.") 
        else:
            pass
        
        if data_source == "gwosc-cvmfs":
            dataset = self.read_gwdata_gwosc_cvmfs(start, end, **kwargs)
        elif data_source == "local":
            dataset = self.read_from_virgo(start=start, end=end, **kwargs)       
            
        elif data_source == "gwosc-online":
            dataset = self.read_gwdata_gwosc_remote(start, end, **kwargs)
        else:
            raise RuntimeError("Data source %s not yet implemented" % data_source)
         
        # Check wether the `dataset` is a gwpy TimeSeries or TimeSeriesDict:        
        if isinstance(dataset, TimeSeriesDict):
            grp = self.create_group(dts_key)
            for k in dataset.keys():
                dset = grp.create_dataset(k, data= dataset[k].data)
                dset.attrs.create('t0', dataset[k].t0)
                dset.attrs.create('unit', str(dataset[k].unit ))
                dset.attrs.create('channel', str(dataset[k].channel))
                dset.attrs.create('sample_rate', dataset[k].sample_rate.value)

        elif isinstance(dataset, TimeSeries):
            dset = self.create_dataset(dts_key, data= dataset.data)
            dset.attrs.create('t0', dataset.t0)
            dset.attrs.create('unit', str(dataset.unit ))
            dset.attrs.create('channel', str(dataset.channel))
            dset.attrs.create('sample_rate', dataset.sample_rate.value)
            
        if return_output:
            return self[dts_key]
                                     
    @staticmethod
    def _search_cvmfs(start=None, end=None,ifo='V1', rate='4k', data_format='hdf5',
                      cvmfs_path='/data2/cvmfs/gwosc.osgstorage.org/gwdata/'):
        """This methood searches the files corresponding to a given format in the
        provided gps time interval, science run (S5, S6, O1 or O2) and interferometer.

        In practice this is a twin of the ``find_gwf`` class method, meant to be
        used primarily on PC-Universe2, where cvmfs is mounted in the provided path,
        or on any other machine changing the path to cvmfs accordignly.

        Parameters
        ----------
        start : `~gwpy.time.LIGOTimeGPS`, float, str, optional
            starting gps time where to find the frame files.
            Default: 1126259462.4 for O1, 1187008882.4 for O2,
            Error for any other choice of run.
            
        end : `~gwpy.time.LIGOTimeGPS`, float, str, optional
            final gps time where to find the frame files. If ``start``
            is not provided, this is automatically set to 60 seconds.
        
        ifo : str, optional
            Name of the interferometer to read the data. Available options: 
                H1, L1, or V1 (default, only for the last part of O2)
 
        data_format : str, optional
            Either ``'hdf5'`` or ``'gwf'``
 
        cvmfs_path : str, optional
            Directory where to look for the hdf5 files. Default choice is the
            directory on PC-Universe2. Don't change this value unless you have
            your own cvmfs on your local machine.
     
        Returns
        -------
        gwf_paths : list of str 
            List of paths to the gwf/hdf5 file(s) in the provided interval.
            
        Notes
        -----
        Still in development.
        """
                
        # 0) Initialise some values if not provided
        if (m_tstart_gps is None):
            m_tstart_gps = 1126259462.4 - 30 # GW150914 - 30 seconds
            m_tstop_gps = m_tstart_gps + 60          # Overwrite the previous 'end'
            warn('Warning! No gps_start provided. GW150914 - 30 sec chosen instead.')
        else:
            m_tstart_gps = to_gps(m_tstart_gps)
            
        if m_tstop_gps is None:
            m_tstop_gps = m_tstart_gps+60
            warn('Warning! No gps_end provided. Duration set to 60 seconds.')     
        else:
            m_tstop_gps = to_gps(m_tstop_gps)
       
        run = find_run(m_tstart_gps, m_tstop_gps)
        
        # Check format and rate
        if m_data_format not in ('hdf5', 'gwf'):
            raise ValueError("Error!! Invalid data format!! It must be either hdf5 or gwf")
        elif m_data_format=='hdf5':
            frmtdir='hdf'
        elif m_data_format=='gwf':
            frmtdir='frame'
        
        if rate not in ('4k', '16k'):
            raise ValueError("Error!! Invalid rate! It must be either 4k or 16k. Remember that you can resample it later.")
        
        # 1) Define the path where to search the files               
        hdf5_paths = glob(join(cvmfs_path,run,'strain.'+rate,frmtdir+'.v1/',ifo,'**/*.'+m_data_format))
        
        # 2) Make a dictionary with start gps time as key, and path, duration, and end as vlas.
        gwf_dict = {int(float(k.split('-')[-2])): {'path': k,
                                                   'len':  int(float(k.split('-')[-1].rstrip(m_data_format))),
                                                    'stop': int(float(k.split('-')[-2]) + float(k.rstrip(m_data_format).split('-')[-1]))
                                                   }
                    for k in hdf5_paths}

        # 3) 
        mindict = {k: v for k, v in gwf_dict.items() if k <= m_tstart_gps}
        try:
            minvalue = max(mindict.keys())
        except ValueError:
            raise ValueError("ERROR!! No GWF file found. Provided gps_start is before the beginning of the ffl.")            
        maxdict = {k: v for k, v in gwf_dict.items() if k >= m_tstart_gps and m_tstop_gps <= v['stop']}
        try:
            maxvalue = min(maxdict.keys())
        except ValueError:
            raise ValueError("ERROR!! No GWF file found. Select new gps time interval or try another frame.")            
            
        gwf_paths = [v['path'] for k, v in gwf_dict.items() if k >= minvalue and k <= maxvalue and v['path'].endswith("."+m_data_format)]

        return gwf_paths
    
    def read_gwdata_gwosc_cvmfs(self, start=None, end=None, ifo='V1',
                                channels=None, nproc=1, rate='4k', data_format='hdf5',
                                crop=True, cvmfs_path='/data2/cvmfs/gwosc.osgstorage.org/gwdata/', **kwargs):
        """
        This method search GW data in the cvmfs aschive with the `_search_cvmfs` method,
        and returns a dataset (either TimeSeries or TimeSeriesDict).
        
        Parameters
        ----------
        run : str, optional
            Name of the run where to read the data. Available options: 
            S5, S6, O1, or O2 (default)
        
        ifo : str, optional
            Name of the interferometer to read the data. Available options: 
            H1, L1, or V1 (default, only for the last part of O2)
        
        cvmfs_path : str, optional
            Directory where to look for the hdf5 files. Default choice is the
            directory on PC-Universe2. Don't change this value unless you have
            your own cvmfs on your local machine.
        
        start : `~gwpy.time.LIGOTimeGPS`, float, str, optional
            starting gps time where to find the frame files. Default: 10 seconds ago
            
        end : `~gwpy.time.LIGOTimeGPS`, float, str, optional
            ending gps time where to find the frame file. If `m_tstart_gps` is not provided, and the default
            value of 10 secods ago is used instead, `mtstop_gps` becomes equal to `m_tstart_gps`+5. If `m_tstart_gps` is
            provided but not `mtstop_gps`, the default duration is set to 5 seconds as well
                   
        nproc : int, optional
            Number of precesses to use for reading the data. This number should be smaller than
            the number of threads that the machine is hable to handle. Also, remember to
            set it to 1 if you are calling this method inside some multiprocessing function
            (otherwise you will spawn an 'army of zombie'. Google it). The best performances
            are obtained when the number of precesses equals the number of gwf files from to read from.
            
        crop : bool, optional
            For some purpose, it can be useful to get the whole content of the gwf files
            corresponding to the data of interest. In this case, set this parameter as False.
            Otherwise, if you prefer the data cropped accordingly to the provided gps interval
            leave it True.
            
        Returns
        -------
        outGwDM : TimeSeriesDict 
            filled with the Datasets corresponding to the specifications
            provided in the input parameters
            
        """    
        # -- Improvement ---------------------
        # It is a waste of lines of code to repeat the gps checks etc. in each  method.
        # This will be improved in a next release 
        
       # 0) Initialise some values if not provided
        if not start:
            start = 1126259462.4 - 30 # GW150914 - 30 seconds
            end = start + 60          # Overwrite the previous 'end'
            warn('Warning! No gps_start provided. GW150914 - 30 sec chosen instead.')

        if not end:
            end = start+60
            warn('Warning! No gps_end provided. Duration set to 60 seconds.')     
    
        run = find_run(start,end)
            
        # Check format and rate
        if data_format not in ('hdf5', 'gwf'):
            raise ValueError("Error!! Invalid data format!! It must be either hdf5 or gwf")
        if rate not in ('4k', '16k'):
            raise ValueError("Error!! Invalid rate! It must be either 4k or 16k. Remember that you can resample it later.")
        
        # If the name of the channel(s) has been provided, replace the 'ifo' with one matching
        if isinstance(channels,list):
            ifo = channels[0][:2]
        elif isinstance(channels,str):
            ifo = channels[:2]
        else:
            # Define the channel names
            if (data_format=='gwf') and (rate=='4k') and (run not in ('O2','O3a')):
                channels=[f'{ifo}:LOSC-STRAIN',f'{ifo}:LOSC-DQMASK' ,f'{ifo}:LOSC-INJMASK']  # 4k before O2
            elif (data_format=='gwf') and (rate=='4k'):
                channels=[f'{ifo}:GWOSC-4KHZ_R1_STRAIN',f'{ifo}:GWOSC-4KHZ_R1_DQMASK' ,f'{ifo}:GWOSC-4KHZ_R1_INJMASK'] # 4k since O2
            elif (data_format=='gwf') and (rate=='16k'):
                channels=[f'{ifo}:GWOSC-16KHZ_R1_STRAIN',f'{ifo}:GWOSC-16KHZ_R1_DQMASK' ,f'{ifo}:GWOSC-16KHZ_R1_INJMASK'] 
        
        # 1) Make a dictionary of hdf5 file paths corresponding to the provided specs

        # Find the paths to the gwf's
        pths = self._search_cvmfs(ifo=ifo, start=start, end=end,
                                  cvmfs_path=cvmfs_path, rate=rate, data_format=data_format)
        
        # If data are read from just one gwf/hdf5, crop it immediately
        if len(pths)==1 and do_crop:
            if data_format=='hdf5':
                out_dataset = TimeSeries.read(source=pths, start=start, end=end, nproc=nproc, format='hdf5.losc', **kwargs)
            elif m_data_format=='gwf':
                out_dataset = TimeSeriesDict.read(source=pths, channels=channels, start=start, end=end,
                                                  nproc=nproc, dtype=np.float64, **kwargs)               
        elif not do_crop:
            if m_data_format=='hdf5':
                out_dataset = TimeSeries.read(source=pths, nproc=nproc, format='hdf5.losc', **kwargs)
            elif m_data_format=='gwf':
                out_dataset = TimeSeriesDict.read(source=pths, channels=channels, nproc=nproc, dtype=np.float64, **kwargs)                
                           
        elif len(pths)>1:
            if data_format=='hdf5':
                try:
                    TS = TimeSeries.read(source=pths, nproc=nproc, format='hdf5.losc', **kwargs)
                    out_dataset = TS.crop(start=to_gps(start), end=to_gps(end))
                except ValueError:
                    warn('WARNING!!! Selected data interval contains holes (missing data). Trying to fetch them form online gwosc' )
                    # Try to get it from online
                    try:
                        out_dataset = TimeSeries.fetch_open_data(ifo, start=start, end=end,
                                                                 tag=kwargs.get('tag','CLN'),format='hdf5', **kwargs)
                        out_dataset.sample_rate = sample_rate
                    except:
                        print('No way of obtaining this data... Sorry!' )
                        out_dataset = TimeSeries([])
                        pass
                
            elif data_format=='gwf':
                try:
                    out_dataset = TimeSeriesDict.read(source=pths, channels=channels, nproc=nproc, dtype=np.float64, **kwargs)
                    out_dataset = out_dataset.crop(start=to_gps(start), end=to_gps(end))
                except ValueError:
                    warn('WARNING!!! Selected data interval contains holes (missing data). Trying to fetch them form online gwosc' )
                    # Try to get it from online
                    try:
                        out_dataset = TimeSeries.fetch_open_data(ifo, start=start, end=end,
                                                                 tag=kwargs.get('tag','CLN'),format='hdf5', **kwargs)
                        out_dataset.sample_rate = sample_rate
                    except:
                        print('No way of obtaining this data... Sorry!' )
                        out_dataset = TimeSeries([])
                        pass           

        else:
            # Return whole frame files: k*100 seconds of data
            if data_format=='hdf5':
                out_dataset = TimeSeries.read(source=pths, start=start, format='hdf5.losc', **kwargs)
            elif data_format=='gwf':            
                out_dataset = TimeSeriesDict.read(source=pths, channels=channels, nproc=nproc, dtype=np.float64, **kwargs)            
              
        return out_dataset
    
    @staticmethod
    def read_gwdata_gwosc_remote(start, end, ifo='V1', data_format="hdf5", rate='4k', **kwargs):
        """
        Read GWOSC data from remote host server, which is by default: host='https://www.gw-openscience.org'
        This method is based on GWpy ``fetch_open_data``
        
        Parameters
        ----------
        ifo : str
            Either ``'L1'``, ``'H1'`` or ``'V1'``, the two-character prefix of the IFO in which you are interested
        start : LIGOTimeGPS, float, str, optional
            Starting gps time where to find the frame files. Default: 10 seconds ago
        end : 
            Stop
        name : str
            Name to give to this dataset
        data_format : hdf5 or gwf
            Data format
        **kwargs
            Any other keyword arguments are passed to the ``TimeSeries.fetch_open_data`` method of GWpy. Refer to `the documentation <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries.fetch_open_data>`_ for more details
            
        Returns
        -------
        gwpy.timeseries.TimeSeries
        
        """
        if rate in ('4k', 4096):
            sample_rate = 4096
        elif rate in ('16k', 16384):
            sample_rate = 16384
        else:
            raise ValueError("Inconsistent 'rate' parameter for gwosc!!! It must be either '4k' or '16k'.")
                 
        TS = TimeSeries.fetch_open_data(ifo, start, end, format=data_format,
                                        sample_rate=sample_rate, **kwargs)     
        TS.sample_rate = sample_rate
        return TS       

    def write_gwdama(self, filename=None, compression="gzip"):
        """
        Method to write the dataset into an hdf5 file. It preserves the hierarchical structure, of course, and all the attributes as metadata.
        
        Parameters
        ----------
        filename : str, optional
            Name of the output file. Default: ``output_gwdama_{}.h5'.format(self.['time_stamp'])``
        compression : str, optional
            Compression level. Default ``'gzip'``. Refer to the `Compression filter documentation <https://docs.h5py.org/en/stable/high/dataset.html#lossless-compression-filters>`_ of ``h5py`` for further details.

        """

        # defaut name
        m_ds = {}
        if not filename:
            filename = 'output_gwdama_{}.h5'.format(self.attrs['time_stamp'])

        if isfile(filename):
            warn('WARNING!! File already exists.')
        creation_time = str(time.strftime("%y-%m-%d_%Hh%Mm%Ss", time.localtime()))

        with h5py.File(filename, 'w') as out_hf:
            out_hf.attrs["time_stamp"] = creation_time
            for a, val in self.attrs.items():
                if a != "time_stamp":
                    out_hf.attrs[a] = val
            for ki in self.keys():
                self.copy(ki,out_hf)

    @staticmethod
    def find_gwf(start=None, end=None, ffl_spec="V1raw", ffl_path=None, gwf_path=None):
        """Fetch and return a list of GWF file paths corresponding to the provided gps time interval.
        Loading all the gwf paths of the data stored at Virgo is quite time consuming. This should be
        done only the first time though. Anyway, it is impossible to be sure that all the paths
        are already present in the class attribute gwf_paths wihout loading them again and checking if
        they are present. This last part should be improved in order to speed up the data acquisition. 
        
        Parameters
        ----------
        start : LIGOTimeGPS, float, str, optional
            starting gps time where to find the frame files. Default: 10 seconds ago
            
        end : LIGOTimeGPS, float, str, optional
            ending gps time where to find the frame file. If ``start`` is not provided, and the default
            value of 10 secods ago is used instead, `end` becomes equal to ``start``+5. If ``start`` is
            provided but not ``end``, the default duration is set to 5 seconds as well
            
        ffl_spec : str, optional
            Pre-encoded specs of the ffl file to read. Available options are: ``V1raw`` (default) for Virgo raw data on farm, ``V1trend``, for data sampled at 1Hz on farm, ``V1trend100`` for 0.01 Hz data on farm, ``H1`` and ``L1`` on farm, ``V1O3a``, ``H1O3a`` and ``L1O3a`` archived from O3a, ``Unipi_arch`` and ``Unipi_O3`` on Unipi servers. The latter are reachable only from "farmrda1" machines
            
        ffl_path : str, optional
            Alternative to  ``ffl_specs``: if the path to a local ffl file is available, the gwf corresponding to the specified
            gps interval are read from it and ``ffl_specs`` is ignored.                  
            
        gwf_path : str, optional
            Altenative to ffl files. It provides a path to a repository (also nested ones) where to look for gwf files. 
            
        Returns
        -------
        gwf_paths : `list`
            List of paths to the gwf file in the provided interval.
            
        """
        data_format = 'gwf'
        
        # 0) Initialise some values if not provided
        if not start:
            start = to_gps('now')-10
            end = start + 5          # Overwrite the previous 'end'
            warn('Warning! No gps_start provided. Changed to 10 seconds ago, and 5 seconds of duration.')
        else:
            start = to_gps(start)
            
        if not end:
            end = start+5
            warn('Warning! No gps_end provided. Duration set to 5 seconds.')     
        else:
            end = to_gps(end)
        
        # Get a list of gwf files. This can be done searching a path or reading an ffl file
        # A) Create a list from the gwf path
        if gwf_path:
            gwf_list = glob(join(gwf_path,f'*.{data_format}'))        
        
            gwf_dict = {int(basename(k).split('-')[-2]): {'path': k,
                                                         'len': int(splitext(basename(k).split('-')[-1])[0]),
                                                         'stop': int(splitext(basename(k).split('-')[-1])[0]) + int(basename(k).split('-')[-2])}
                        for k in gwf_list}
        # B) Obtain the list from ffl files
        else:
            if ffl_path:
                ffl = ffl_path
            else:            
                # Find where to fetch the data
                # Virgo
                if ffl_spec=="V1raw":
                    ffl = ffl_paths.V1_raw
                elif ffl_spec=="V1trend":
                    ffl = ffl_paths.V1_trend
                elif ffl_spec=="V1trend100":
                    ffl = ffl_paths.V1_trend100
                # at UniPi
                elif ffl_spec=="Unipi_O3":
                    ffl = ffl_paths.unipi_O3
                elif ffl_spec=="Unipi_arch":
                    ffl = ffl_paths.unipi_arch

                # O3a
                elif ffl_spec=="V1O3a":
                    ffl = ffl_paths.V1_O3a
                elif ffl_spec=="H1O3a":
                    ffl = ffl_paths.H1_O3a
                elif ffl_spec=="L1O3a":
                    ffl = ffl_paths.L1_O3a    

                # LIGO data from CIT
                # >>>> FIX: multiple frame <<<<
                elif ffl_spec=="H1":
                    if start<(to_gps('now')-3700):
                        ffl = ffl_paths.H1_older
                    elif start>(to_gps('now')-3600):
                        ffl = ffl_paths.H1_newer
                    else:
                        # <------ Fix reading from eterogeneous frames ----->
                        warn("Warning!!! Data not available online and not stored yet")
                elif ffl_spec=="L1":
                    if end<(to_gps('now')-3700):
                        ffl = ffl_paths.L1_older
                    elif end>(to_gps('now')-3600):
                        ffl = ffl_paths.L1_newer
                    else:
                        # <------ Fix reading from eterogeneous frames ----->
                        warn("Warning!!! Data not available online and not stored yet")        
                else:
                    raise ValueError("ERROR!! Unrecognised ffl spec. Check docstring")            

            # 1) Get the ffl (gwf list) corresponding to the desired data frame
            with open(ffl, 'r') as f:
                content = f.readlines()
                # content is a list (with hundreds of thousands of elements) of strings
                # containing the path to the gwf, the gps_start, the duration, and other
                # floats, usually equals to zero.
            content = [x.split() for x in content]

            # Make a dictionary with start gps time as key, and path, duration, and end as vlas.
            gwf_dict = {round(float(k[1])): {'path': k[0],
                                             'len': int(float(k[2])),
                                             'stop': round(float(k[1]) + int(float(k[2])))}
                        for k in content}

        # 2)
        mindict = {k: v for k, v in gwf_dict.items() if k <= start}
        try:
            minvalue = max(mindict.keys())
        except ValueError:
            raise ValueError("ERROR!! No GWF file found. Provided gps_start is before the beginning of the ffl.")            
        maxdict = {k: v for k, v in gwf_dict.items() if k >= start and end <= v['stop']}
        try:
            maxvalue = min(maxdict.keys())
        except ValueError:
            raise ValueError("ERROR!! No GWF file found. Select new gps time interval or try another frame.")            
            
        gwf_paths = [v['path'] for k, v in gwf_dict.items() if k >= minvalue and k <= maxvalue and v['path'].endswith(f'.{data_format}')]

        return gwf_paths
    
    @classmethod
    def read_from_virgo(cls, channels, start=None, end=None, nproc=1, crop=True, **kwargs):
        """This method reads GW data from GWFs, fetched with the ``ffind_gwf`` method,
        and returns a TimeSeriesDictionary object filled with data.
        
        Parameters
        ----------
        channels : list of strings
            List with the names of the Virgo channels whose data that should be read.
            Example: channels = ['V1:Hrec_hoft_16384Hz', 'V1:LSC_DARM']
        
        start : LIGOTimeGPS, float, str, optional
            starting gps time where to find the frame files. Default: 10 seconds ago
            
        end : LIGOTimeGPS, float, str, optional
            ending gps time where to find the frame file. If `start` is not provided, and the default
            value of 10 secods ago is used instead, `end` becomes equal to `start`+5. If `start` is
            provided but not `end`, the default duration is set to 5 seconds as well
                   
        nproc : int, optional
            Number of precesses to use for reading the data. This number should be smaller than
            the number of threads that the machine is hable to handle. Also, remember to
            set it to 1 if you are calling this method inside some multiprocessing function
            (otherwise you will spawn an 'army of zombie'. Google it). The best performances
            are obtained when the number of precesses equals the number of gwf files from to read from.
            
        crop : bool, optional
            For some purpose, it can be useful to get the whole content of the gwf files
            corresponding to the data of interest. In this case, set this parameter as False.
            Otherwise, if you prefer the data cropped accordingly to the provided gps interval
            leave it True.
            
        kwargs : dict, optional
            It contains keyword arguments to pass to ``find_gwf`` method, in particular ``ffl_spec``, ``ffl_path``, or ``gwf_path``
            and those for ``TimeSeries.read(...)``. Refer to the corresponding documentation for more details.
            
        Returns
        -------
        outTSD : TimeSeriesDict object
            Dictionary of TimeSeries corresponding to the specifications provided in the input parameters.
        """

        if isinstance(channels, str):
            channels = [channels]
        
        # Find the paths to the gwf's
        pths = cls.find_gwf(start=start, end=end, ffl_spec=kwargs.get("ffl_spec","V1raw"),
                            ffl_path=kwargs.get('ffl_path'), gwf_path=kwargs.get('gwf_path'))
        
        kwargs.pop('ffl_spec',None)
        kwargs.pop('ffl_path',None)
        kwargs.pop('gwf_path',None)
        
        # If data are read from just one gwf, crop it immediately
        if len(pths)==1 and crop:
            outTSD = TimeSeriesDict.read(source=pths, channels=channels, start=start, end=end,
                            nproc=nproc, dtype=np.float64, **kwargs)
        elif not crop:
            outTSD = TimeSeriesDict.read(source=pths, channels=channels, nproc=nproc, dtype=np.float64, **kwargs)

        elif len(pths)>1:
            outTSD = TimeSeriesDict.read(source=pths, channels=channels, nproc=nproc, dtype=np.float64, **kwargs)
            outTSD = outTSD.crop(start=to_gps(start), end=to_gps(end))
        else:
            # Return whole frame files: k*100 seconds of data
            outTSD = TimeSeriesDict.read(source=pths, channels=channels, nproc=nproc, dtype=np.float64, **kwargs)

        if len(outTSD)==1:
            outTSD = next(iter(outTSD.values()))
        return outTSD

    
    def from_TimeSeries(self, TS, dts_key=None, **kwgs):
        """
        Import into the current instance of :class:`~gwdama.io.GwDataManager` the TimeSeries or TimeSeriesDict of GWpy passed as an argument ot this method.
        
        Parameters
        ----------
        TS : `TimeSeries <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_ or `TimeSeriesDiuct <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeriesDict>`_
            GWpy object to import in :ref:`index:GWdama`
        dts_key : str, optional
            If ``TS`` is an instance of `TimeSeries <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_, this parameter corresponds to the name to give to the :class:`~gwdama.io.Dataset` added to the current instance of :class:`~gwdama.io.GwDataManager`. If ``TS`` is a `TimeSeriesDiuct <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeriesDict>`_, this is the name of the group where to create a :class:`~gwdama.io.Dataset` for every item of ``TS``. In this case, their names will be the keys of ``TS``
        
        **kwgs : dict, optional
            Other parameters that can be passed to the `.create_dataset method of h5py <https://docs.h5py.org/en/stable/high/group.html?highlight=create_dataset#h5py.Group.create_dataset>`_. This includes details on the compression format of this data: *e.g.* ``compression="gzip", compression_opts=9``. Refer to the documentation of h5py for further details
        
        Returns
        -------
         : Dataset or Group
            Dataset or Group corresponding to the imported `TimeSeries <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeries>`_ or `TimeSeriesDiuct <https://gwpy.github.io/docs/stable/api/gwpy.timeseries.TimeSeries.html#gwpy.timeseries.TimeSeriesDict>`_

        See Also
        --------
        gwdama.io.Dataset.to_TimeSeries
        """
        
        if isinstance(TS, TimeSeriesDict):
            dts_key = dts_key or 'TS'
            for k, ts in TS.items():
                self.from_TimeSeries(ts, dts_key=f"{dts_key}/{k}", **kwgs)
            
        elif isinstance(TS, TimeSeries):
            dts_key = dts_key or 'TS'
            if dts_key in self:
                raise ValueError("Provided 'dts_key' already present in this dama. Please, pass a new name for it")
            self.create_dataset(name=dts_key, data=TS.value, **kwgs)
            
            # Add some attributes
            dts = self[dts_key]
            dts.attrs['t0'] = TS.t0.value
            dts.attrs['sample_rate'] = TS.sample_rate.value
            dts.attrs['unit'] = str(TS.unit)
            try:
                dts.attrs['channel'] = TS.channel.value
            except AttributeError:
                dts.attrs['channel'] = ''

        return self[dts_key]

#     def from_DataFrame(self, df, dts_key=None, **kwgs):
#         """
#         WORK in progress
        
#         Import into the current instance of :class:`~gwdama.io.GwDataManager` 
 
#         """
#         pass
        
#         if isinstance(TS, TimeSeriesDict):
#             dts_key = dts_key or 'TS'
#             for k, ts in TS.items():
#                 self.from_TimeSeries(ts, dts_key=f"{dts_key}/{k}", **kwgs)
            
#         elif isinstance(TS, TimeSeries):
#             dts_key = dts_key or 'TS'
#             if dts_key in self:
#                 raise ValueError("Provided 'dts_key' already present in this dama. Please, pass a new name for it")
#             self.create_dataset(name=dts_key, data=TS.value, **kwgs)
            
#             # Add some attributes
#             dts = self[dts_key]
#             dts.attrs['t0'] = TS.t0.value
#             dts.attrs['sample_rate'] = TS.sample_rate.value
#             dts.attrs['unit'] = str(TS.unit)
#             try:
#                 dts.attrs['channel'] = TS.channel.value
#             except AttributeError:
#                 dts.attrs['channel'] = ''

#         return self[dts_key]



# This file tests the reading capabilities
# using gwosc datasets

import unittest
from os import getcwd, path 
import sys
from pandas import read_csv
from pandas.testing import assert_frame_equal
from gwdama.io import GwDataManager

class TestDama(unittest.TestCase):
    """
    Class to test if the output af a simple data acquisition with gwdama performs as expected
    """

    def setUp(self):
        """
        This method will be called before the test and reads the existing reference DataFrame
        """
        df_name = "glitch_sample.csv"
        script_path = path.join(getcwd(), path.dirname(sys.argv[0]))
        data_dir = path.join(path.dirname(script_path),"..","data")
        
        try:
            ref_df = read_csv(path.join(data_dir, df_name))
        except IOError:
            print(f"Cannot access the reference glitch Dataframe at {data_dir}")      
        # Store the reference DataFrame
        self.ref_df = ref_df
      
        gl_csv = "glitch_sample_list_L1_O2.csv"
        try:
            gl_df = read_csv(path.join(data_dir, gl_csv))
        except IOError:
            print(f"Cannot access glitch specs Dataframe at {data_dir}")
        
        ifo = gl_df['ifo'].unique()[0]

        # Instance GwDataManager class
        gldama = GwDataManager("test_glitch_gwosc")

        # Import glitches data
        for n, t0_gps in enumerate(gl_df['GPStime']):
            t1=t0_gps-.5
            t2=t0_gps+.5
            gldama.read_gwdata(t1, t2, ifo=ifo, dts_key=f"Gltich_{n+1}")
        
        # Convert to DataFrame and store
        self.test_df = gldama.to_DataFrame()
   
    def test_dama(self):
        """
        Test that the reference DataFrame in data is equal to the acquired one
        """
        #foo = pd.DataFrame()
        assert_frame_equal(self.ref_df, self.test_df)

if __name__ == '__main__':
    unittest.main()          